import duckdb
import datetime
import numpy as np
import pytest
import decimal
import math
from decimal import Decimal
import re
from conftest import NumpyPandas, ArrowPandas

standard_vector_size = duckdb.__standard_vector_size__


def create_generic_dataframe(data, pandas):
    return pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})


def create_repeated_nulls(size):
    data = [None, "a"]
    n = size
    data = data * n
    return data


def create_trailing_non_null(size):
    data = [None for _ in range(size - 1)]
    data.append('this is a long string')
    return data


class IntString:
    def __init__(self, value: int):
        self.value = value

    def __str__(self):
        return str(self.value)


# To avoid DECIMAL being upgraded to DOUBLE (because DOUBLE outranks DECIMAL as a LogicalType)
# These floats had their precision preserved as string and are now cast to decimal.Decimal
def ConvertStringToDecimal(data: list, pandas):
    for i in range(len(data)):
        if isinstance(data[i], str):
            data[i] = decimal.Decimal(data[i])
    data = pandas.Series(data=data, dtype='object')
    return data


class TestResolveObjectColumns(object):
    # TODO: add support for ArrowPandas
    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_integers(self, pandas, duckdb_cursor):
        data = [5, 0, 3]
        df_in = create_generic_dataframe(data, pandas)
        # These are float64 because pandas would force these to be float64 even if we set them to int8, int16, int32, int64 respectively
        df_expected_res = pandas.DataFrame({'0': pandas.Series(data=data, dtype='int32')})
        df_out = duckdb_cursor.sql("SELECT * FROM df_in").df()
        print(df_out)
        pandas.testing.assert_frame_equal(df_expected_res, df_out)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_correct(self, pandas, duckdb_cursor):
        data = [{'a': 1, 'b': 3, 'c': 3, 'd': 7}]
        df = pandas.DataFrame({'0': pandas.Series(data=data)})
        duckdb_col = duckdb_cursor.sql("SELECT {a: 1, b: 3, c: 3, d: 7} as '0'").df()
        converted_col = duckdb_cursor.sql("SELECT * FROM df").df()
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_fallback_different_keys(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'e': 7}],  #'e' instead of 'd' as key
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )

        converted_df = duckdb_cursor.sql("SELECT * FROM x").df()
        y = pandas.DataFrame(
            [
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'e'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
            ]
        )
        equal_df = duckdb_cursor.sql("SELECT * FROM y").df()
        pandas.testing.assert_frame_equal(converted_df, equal_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_fallback_incorrect_amount_of_keys(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3}],  # incorrect amount of keys
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )
        converted_df = duckdb_cursor.sql("SELECT * FROM x").df()
        y = pandas.DataFrame(
            [
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c'], 'value': [1, 3, 3]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
            ]
        )
        equal_df = duckdb_cursor.sql("SELECT * FROM y").df()
        pandas.testing.assert_frame_equal(converted_df, equal_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_value_upgrade(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'a': 1, 'b': 3, 'c': 3, 'd': 'string'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )
        y = pandas.DataFrame(
            [
                [{'a': 1, 'b': 3, 'c': 3, 'd': 'string'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': '7'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': '7'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': '7'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': '7'}],
            ]
        )
        converted_df = duckdb_cursor.sql("SELECT * FROM x").df()
        equal_df = duckdb_cursor.sql("SELECT * FROM y").df()
        pandas.testing.assert_frame_equal(converted_df, equal_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_null(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [None],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )
        y = pandas.DataFrame(
            [
                [None],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )
        converted_df = duckdb_cursor.sql("SELECT * FROM x").df()
        equal_df = duckdb_cursor.sql("SELECT * FROM y").df()
        pandas.testing.assert_frame_equal(converted_df, equal_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_fallback_value_upgrade(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'a': 1, 'b': 3, 'c': 3, 'd': 'test'}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
                [{'a': 1, 'b': 3, 'c': 3, 'd': 7}],
            ]
        )
        y = pandas.DataFrame(
            [
                [{'a': '1', 'b': '3', 'c': '3', 'd': 'test'}],
                [{'a': '1', 'b': '3', 'c': '3', 'd': '7'}],
                [{'a': '1', 'b': '3', 'c': '3'}],
                [{'a': '1', 'b': '3', 'c': '3', 'd': '7'}],
                [{'a': '1', 'b': '3', 'c': '3', 'd': '7'}],
            ]
        )
        converted_df = duckdb_cursor.sql("SELECT * FROM x").df()
        equal_df = duckdb_cursor.sql("SELECT * FROM y").df()
        pandas.testing.assert_frame_equal(converted_df, equal_df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_correct(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
            ]
        )
        x.rename(columns={0: 'a'}, inplace=True)
        converted_col = duckdb_cursor.sql("select * from x as 'a'").df()
        duckdb_cursor.sql(
            """
            CREATE TABLE tmp(
                a MAP(VARCHAR, INTEGER)
            );
        """
        )
        for _ in range(5):
            duckdb_cursor.sql(
                """
                INSERT INTO tmp VALUES (MAP(['a', 'b', 'c', 'd'], [1, 3, 3, 7]))
            """
            )
        duckdb_col = duckdb_cursor.sql("select a from tmp AS '0'").df()
        print(duckdb_col.columns)
        print(converted_col.columns)
        pandas.testing.assert_frame_equal(converted_col, duckdb_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    @pytest.mark.parametrize('sample_size', [1, 10])
    @pytest.mark.parametrize('fill', [1000, 10000])
    @pytest.mark.parametrize('get_data', [create_repeated_nulls, create_trailing_non_null])
    def test_analyzing_nulls(self, pandas, duckdb_cursor, fill, sample_size, get_data):
        data = get_data(fill)
        df1 = pandas.DataFrame(data={"col1": data})
        duckdb_cursor.execute(f"SET GLOBAL pandas_analyze_sample={sample_size}")
        df = duckdb_cursor.execute("select * from df1").df()

        pandas.testing.assert_frame_equal(df1, df)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_nested_map(self, pandas, duckdb_cursor):
        df = pandas.DataFrame(data={'col1': [{'a': {'b': {'x': 'A', 'y': 'B'}}}, {'c': {'b': {'x': 'A'}}}]})

        rel = duckdb_cursor.sql("select * from df")
        expected_rel = duckdb_cursor.sql(
            """
            select x::map(varchar, struct(b map(varchar, varchar))) col1 from (VALUES
                ('{a={b: {x=A, y=B}}}'),
                ('{c={b: {x=A}}}')
            ) t(x)
        """
        )

        res = str(rel)
        expected_res = str(expected_rel)
        assert res == expected_res

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_value_upgrade(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 'test']}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
                [{'key': ['a', 'b', 'c', 'd'], 'value': [1, 3, 3, 7]}],
            ]
        )
        x.rename(columns={0: 'a'}, inplace=True)
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql(
            """
            CREATE TABLE tmp2(
                a MAP(VARCHAR, VARCHAR)
            );
        """
        )
        duckdb_cursor.sql(
            """
            INSERT INTO tmp2 VALUES (MAP(['a', 'b', 'c', 'd'], ['1', '3', '3', 'test']))
        """
        )
        for _ in range(4):
            duckdb_cursor.sql(
                """
                INSERT INTO tmp2 VALUES (MAP(['a', 'b', 'c', 'd'], ['1', '3', '3', '7']))
            """
            )
        duckdb_col = duckdb_cursor.sql("select a from tmp2 AS '0'").df()
        print(duckdb_col.columns)
        print(converted_col.columns)
        pandas.testing.assert_frame_equal(converted_col, duckdb_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_duplicate(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([[{'key': ['a', 'a', 'b'], 'value': [4, 0, 4]}]])
        with pytest.raises(
            duckdb.InvalidInputException, match="Dict->Map conversion failed because 'key' list contains duplicates"
        ):
            converted_col = duckdb_cursor.sql("select * from x").df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_nullkey(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([[{'key': [None, 'a', 'b'], 'value': [4, 0, 4]}]])
        with pytest.raises(
            duckdb.InvalidInputException, match="Dict->Map conversion failed because 'key' list contains None"
        ):
            converted_col = duckdb_cursor.sql("select * from x").df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_nullkeylist(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([[{'key': None, 'value': None}]])
        # Isn't actually converted to MAP because isinstance(None, list) != True
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_col = duckdb_cursor.sql("SELECT {key: NULL, value: NULL} as '0'").df()
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_fallback_nullkey(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([[{'a': 4, None: 0, 'c': 4}], [{'a': 4, None: 0, 'd': 4}]])
        with pytest.raises(
            duckdb.InvalidInputException, match="Dict->Map conversion failed because 'key' list contains None"
        ):
            converted_col = duckdb_cursor.sql("select * from x").df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_map_fallback_nullkey_coverage(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{'key': None, 'value': None}],
                [{'key': None, None: 5}],
            ]
        )
        with pytest.raises(
            duckdb.InvalidInputException, match="Dict->Map conversion failed because 'key' list contains None"
        ):
            converted_col = duckdb_cursor.sql("select * from x").df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_key_conversion(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [{IntString(5): 1, IntString(-25): 3, IntString(32): 3, IntString(32456): 7}],
            ]
        )
        duckdb_col = duckdb_cursor.sql("select {'5':1, '-25':3, '32':3, '32456':7} as '0'").df()
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql("drop view if exists tbl")
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_correct(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([{'0': [[5], [34], [-245]]}])
        duckdb_col = duckdb_cursor.sql("select [[5], [34], [-245]] as '0'").df()
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql("drop view if exists tbl")
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_contains_null(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([{'0': [[5], None, [-245]]}])
        duckdb_col = duckdb_cursor.sql("select [[5], NULL, [-245]] as '0'").df()
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql("drop view if exists tbl")
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_starts_with_null(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([{'0': [None, [5], [-245]]}])
        duckdb_col = duckdb_cursor.sql("select [NULL, [5], [-245]] as '0'").df()
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql("drop view if exists tbl")
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_value_upgrade(self, pandas, duckdb_cursor):
        x = pandas.DataFrame([{'0': [['5'], [34], [-245]]}])
        duckdb_col = duckdb_cursor.sql("select [['5'], ['34'], ['-245']] as '0'").df()
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql("drop view if exists tbl")
        pandas.testing.assert_frame_equal(duckdb_col, converted_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_list_column_value_upgrade(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [[1, 25, 300]],
                [[500, 345, 30]],
                [[50, 'a', 67]],
            ]
        )
        x.rename(columns={0: 'a'}, inplace=True)
        converted_col = duckdb_cursor.sql("select * from x").df()
        duckdb_cursor.sql(
            """
            CREATE TABLE tmp3(
                a VARCHAR[]
            );
        """
        )
        duckdb_cursor.sql(
            """
            INSERT INTO tmp3 VALUES (['1', '25', '300'])
        """
        )
        duckdb_cursor.sql(
            """
            INSERT INTO tmp3 VALUES (['500', '345', '30'])
        """
        )
        duckdb_cursor.sql(
            """
            INSERT INTO tmp3 VALUES (['50', 'a', '67'])
        """
        )
        duckdb_col = duckdb_cursor.sql("select a from tmp3 AS '0'").df()
        print(duckdb_col.columns)
        print(converted_col.columns)
        pandas.testing.assert_frame_equal(converted_col, duckdb_col)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_ubigint_object_conversion(self, pandas, duckdb_cursor):
        # UBIGINT + TINYINT would result in HUGEINT, but conversion to HUGEINT is not supported yet from pandas->duckdb
        # So this instead becomes a DOUBLE
        data = [18446744073709551615, 0]
        x = pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})
        converted_col = duckdb_cursor.sql("select * from x").df()
        if pandas.backend == 'numpy_nullable':
            float64 = np.dtype('float64')
            assert isinstance(converted_col['0'].dtype, float64.__class__) == True
        else:
            uint64 = np.dtype('uint64')
            assert isinstance(converted_col['0'].dtype, uint64.__class__) == True

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_double_object_conversion(self, pandas, duckdb_cursor):
        data = [18446744073709551616, 0]
        x = pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})
        converted_col = duckdb_cursor.sql("select * from x").df()
        double_dtype = np.dtype('float64')
        assert isinstance(converted_col['0'].dtype, double_dtype.__class__) == True

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numpy_object_with_stride(self, pandas, duckdb_cursor):
        df = pandas.DataFrame(columns=["idx", "evens", "zeros"])

        df["idx"] = list(range(10))
        for col in df.columns[1:]:
            df[col].values[:] = 0

        counter = 0
        for i in range(10):
            df.loc[df["idx"] == i, "evens"] += counter
            counter += 2

        res = duckdb_cursor.sql("select * from df").fetchall()
        assert res == [
            (0, 0, 0),
            (1, 2, 0),
            (2, 4, 0),
            (3, 6, 0),
            (4, 8, 0),
            (5, 10, 0),
            (6, 12, 0),
            (7, 14, 0),
            (8, 16, 0),
            (9, 18, 0),
        ]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numpy_stringliterals(self, pandas, duckdb_cursor):
        df = pandas.DataFrame({"x": list(map(np.str_, range(3)))})

        res = duckdb_cursor.execute("select * from df").fetchall()
        assert res == [('0',), ('1',), ('2',)]

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_integer_conversion_fail(self, pandas, duckdb_cursor):
        data = [2**10000, 0]
        x = pandas.DataFrame({'0': pandas.Series(data=data, dtype='object')})
        converted_col = duckdb_cursor.sql("select * from x").df()
        print(converted_col['0'])
        double_dtype = np.dtype('object')
        assert isinstance(converted_col['0'].dtype, double_dtype.__class__) == True

    # Most of the time numpy.datetime64 is just a wrapper around a datetime.datetime object
    # But to support arbitrary precision, it can fall back to using an `int` internally

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])  # Which we don't support yet
    def test_numpy_datetime(self, pandas, duckdb_cursor):
        numpy = pytest.importorskip("numpy")

        data = []
        data += [numpy.datetime64('2022-12-10T21:38:24.578696')] * standard_vector_size
        data += [numpy.datetime64('2022-02-21T06:59:23.324812')] * standard_vector_size
        data += [numpy.datetime64('1974-06-05T13:12:01.000000')] * standard_vector_size
        data += [numpy.datetime64('2049-01-13T00:24:31.999999')] * standard_vector_size
        x = pandas.DataFrame({'dates': pandas.Series(data=data, dtype='object')})
        res = duckdb_cursor.sql("select distinct * from x").df()
        assert len(res['dates'].__array__()) == 4

    @pytest.mark.parametrize('pandas', [NumpyPandas()])
    def test_numpy_datetime_int_internally(self, pandas, duckdb_cursor):
        numpy = pytest.importorskip("numpy")

        data = [numpy.datetime64('2022-12-10T21:38:24.0000000000001')]
        x = pandas.DataFrame({'dates': pandas.Series(data=data, dtype='object')})
        with pytest.raises(
            duckdb.ConversionException,
            match=re.escape("Conversion Error: Unimplemented type for cast (BIGINT -> TIMESTAMP)"),
        ):
            rel = duckdb.query_df(x, "x", "create table dates as select dates::TIMESTAMP WITHOUT TIME ZONE from x")

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_fallthrough_object_conversion(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                [IntString(4)],
                [IntString(2)],
                [IntString(0)],
            ]
        )
        duckdb_col = duckdb_cursor.sql("select * from x").df()
        df_expected_res = pandas.DataFrame({'0': pandas.Series(['4', '2', '0'])})
        pandas.testing.assert_frame_equal(duckdb_col, df_expected_res)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal(self, pandas, duckdb_cursor):
        # DuckDB uses DECIMAL where possible, so all the 'float' types here are actually DECIMAL
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                (5,                   5002340,                  -234234234234.0),
                (12.0,                13,                       324234234.00000005),
                (-123.0,              -12.0000000005,           -128),
                (-234234.0,           7453324234.0,             345345),
                (NULL,                NULL,                     1.00000),
                (1.234,               -324234234,               1324234359)
            ) tbl(a, b, c);
        """
        duckdb_cursor.execute(reference_query)
        # Because of this we need to wrap these native floats as DECIMAL for this test, to avoid these decimals being "upgraded" to DOUBLE
        x = pandas.DataFrame(
            {
                '0': ConvertStringToDecimal([5, '12.0', '-123.0', '-234234.0', None, '1.234'], pandas),
                '1': ConvertStringToDecimal(
                    [5002340, 13, '-12.0000000005', '7453324234.0', None, '-324234234'], pandas
                ),
                '2': ConvertStringToDecimal(
                    ['-234234234234.0', '324234234.00000005', -128, 345345, '1E5', '1324234359'], pandas
                ),
            }
        )
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb_cursor.sql("select * from x").fetchall()

        assert conversion == reference

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_coverage(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            {'0': [Decimal("nan"), Decimal("+nan"), Decimal("-nan"), Decimal("inf"), Decimal("+inf"), Decimal("-inf")]}
        )
        conversion = duckdb_cursor.sql("select * from x").fetchall()
        print(conversion[0][0].__class__)
        for item in conversion:
            assert isinstance(item[0], float)
        assert math.isnan(conversion[0][0])
        assert math.isnan(conversion[1][0])
        assert math.isnan(conversion[2][0])
        assert math.isinf(conversion[3][0])
        assert math.isinf(conversion[4][0])
        assert math.isinf(conversion[5][0])
        assert str(conversion) == '[(nan,), (nan,), (nan,), (inf,), (inf,), (inf,)]'

    # Test that the column 'offset' is actually used when converting,

    @pytest.mark.parametrize(
        'pandas', [NumpyPandas(), ArrowPandas()]
    )  # and that the same 2048 (STANDARD_VECTOR_SIZE) values are not being scanned over and over again
    def test_multiple_chunks(self, pandas, duckdb_cursor):
        data = []
        data += [datetime.date(2022, 9, 13) for x in range(standard_vector_size)]
        data += [datetime.date(2022, 9, 14) for x in range(standard_vector_size)]
        data += [datetime.date(2022, 9, 15) for x in range(standard_vector_size)]
        data += [datetime.date(2022, 9, 16) for x in range(standard_vector_size)]
        x = pandas.DataFrame({'dates': pandas.Series(data=data, dtype='object')})
        res = duckdb_cursor.sql("select distinct * from x").df()
        assert len(res['dates'].__array__()) == 4

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_multiple_chunks_aggregate(self, pandas, duckdb_cursor):
        duckdb_cursor.execute(f"SET GLOBAL pandas_analyze_sample=4096")
        duckdb_cursor.execute(
            "create table dates as select '2022-09-14'::DATE + INTERVAL (i::INTEGER) DAY as i from range(4096) tbl(i);"
        )
        rel = duckdb_cursor.sql("select * from dates")
        res = rel.df()
        date_df = res.copy()

        # Convert the dataframe to datetime
        date_df['i'] = pandas.to_datetime(res['i']).dt.date
        assert str(date_df['i'].dtype) == 'object'

        expected_res = [
            (
                1840017600.0,  # Saturday, April 22, 2028 12:00:00 PM (avg)
                1663113600.0,  # Wednesday, September 14, 2022 12:00:00 AM (min)
                2016921600.0,  # Wednesday, November 30, 2033 12:00:00 AM (max)
            )
        ]

        rel = duckdb_cursor.sql(
            """
            select
                avg(epoch(i)),
                min(epoch(i)),
                max(epoch(i))
            from date_df
        """
        )
        actual_res = rel.fetchall()
        assert expected_res == actual_res

        # Now interleave nulls into the dataframe
        duckdb_cursor.execute('drop table dates')
        for i in range(0, len(res['i']), 2):
            res.loc[i, 'i'] = None
        duckdb_cursor.execute('create table dates as select * from res')

        expected_res = [
            (
                1840060800.0,  # Sunday, April 23, 2028 12:00:00 AM
                1663200000.0,  # Thursday, September 15, 2022 12:00:00 AM
                2016921600.0,  # Wednesday, November 30, 2033 12:00:00 AM
            )
        ]
        # Convert the dataframe to datetime
        date_df = res.copy()
        date_df['i'] = pandas.to_datetime(res['i']).dt.date
        assert str(date_df['i'].dtype) == 'object'

        actual_res = duckdb_cursor.sql(
            """
            select
                avg(epoch(i)),
                min(epoch(i)),
                max(epoch(i))
            from date_df
        """
        ).fetchall()

        assert expected_res == actual_res

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_mixed_object_types(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            {
                'nested': pandas.Series(
                    data=[{'a': 1, 'b': 2}, [5, 4, 3], {'key': [1, 2, 3], 'value': ['a', 'b', 'c']}], dtype='object'
                ),
            }
        )
        res = duckdb_cursor.sql("select * from x").df()
        assert res['nested'].dtype == np.dtype('object')

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_deeply_nested_in_struct(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            [
                {
                    # STRUCT(b STRUCT(x VARCHAR, y VARCHAR))
                    'a': {'b': {'x': 'A', 'y': 'B'}}
                },
                {
                    # STRUCT(b STRUCT(x VARCHAR))
                    'a': {'b': {'x': 'A'}}
                },
            ]
        )
        # The dataframe has incompatible struct schemas in the nested child
        # This gets upgraded to STRUCT(b MAP(VARCHAR, VARCHAR))
        res = duckdb_cursor.sql("select * from x").fetchall()
        assert res == [({'b': {'key': ['x', 'y'], 'value': ['A', 'B']}},), ({'b': {'key': ['x'], 'value': ['A']}},)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_struct_deeply_nested_in_list(self, pandas, duckdb_cursor):
        x = pandas.DataFrame(
            {
                'a': [
                    [
                        # STRUCT(x VARCHAR, y VARCHAR)[]
                        {'x': 'A', 'y': 'B'},
                        # STRUCT(x VARCHAR)[]
                        {'x': 'A'},
                    ]
                ]
            }
        )
        # The dataframe has incompatible struct schemas in the nested child
        # This gets upgraded to STRUCT(b MAP(VARCHAR, VARCHAR))
        res = duckdb_cursor.sql("select * from x").fetchall()
        assert res == [([{'key': ['x', 'y'], 'value': ['A', 'B']}, {'key': ['x'], 'value': ['A']}],)]

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_analyze_sample_too_small(self, pandas, duckdb_cursor):
        data = [1 for _ in range(9)] + [[1, 2, 3]] + [1 for _ in range(9991)]
        x = pandas.DataFrame({'a': pandas.Series(data=data)})
        with pytest.raises(duckdb.InvalidInputException, match="Failed to cast value: Unimplemented type for cast"):
            res = duckdb_cursor.sql("select * from x").df()

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_zero_fractional(self, pandas, duckdb_cursor):
        decimals = pandas.DataFrame(
            data={
                "0": [
                    Decimal("0.00"),
                    Decimal("125.90"),
                    Decimal("0.001"),
                    Decimal("2502.63"),
                    Decimal("0.000123"),
                    Decimal("0.00"),
                    Decimal("321.00"),
                ]
            }
        )
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                (0.00),
                (125.90),
                (0.001),
                (2502.63),
                (0.000123),
                (0.00),
                (321.00)
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb.query_df(decimals, "x", "select * from x").fetchall()

        assert conversion == reference

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_incompatible(self, pandas, duckdb_cursor):
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                (5,                   5002340,                  -234234234234),
                (12.0,                13,                       324234234.00000005),
                (-123.0,              -12.0000000005,           -128),
                (-234234.0,           7453324234,               345345),
                (NULL,                NULL,                     0),
                (1.234,               -324234234,               1324234359)
            ) tbl(a, b, c);
        """
        duckdb_cursor.execute(reference_query)
        x = pandas.DataFrame(
            {
                '0': ConvertStringToDecimal(['5', '12.0', '-123.0', '-234234.0', None, '1.234'], pandas),
                '1': ConvertStringToDecimal([5002340, 13, '-12.0000000005', 7453324234, None, '-324234234'], pandas),
                '2': ConvertStringToDecimal(
                    [-234234234234, '324234234.00000005', -128, 345345, 0, '1324234359'], pandas
                ),
            }
        )
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb_cursor.sql("select * from x").fetchall()

        assert conversion == reference
        print(reference)
        print(conversion)

    @pytest.mark.parametrize(
        'pandas', [NumpyPandas(), ArrowPandas()]
    )  # result: [('1E-28',), ('10000000000000000000000000.0',)]
    def test_numeric_decimal_combined(self, pandas, duckdb_cursor):
        decimals = pandas.DataFrame(
            data={"0": [Decimal("0.0000000000000000000000000001"), Decimal("10000000000000000000000000.0")]}
        )
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                (0.0000000000000000000000000001),
                (10000000000000000000000000.0),
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb.query_df(decimals, "x", "select * from x").fetchall()
        assert conversion == reference
        print(reference)
        print(conversion)

    # result: [('1234.0',), ('123456789.0',), ('1234567890123456789.0',), ('0.1234567890123456789',)]
    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_varying_sizes(self, pandas, duckdb_cursor):
        decimals = pandas.DataFrame(
            data={
                "0": [
                    Decimal("1234.0"),
                    Decimal("123456789.0"),
                    Decimal("1234567890123456789.0"),
                    Decimal("0.1234567890123456789"),
                ]
            }
        )
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                    (1234.0),
                    (123456789.0),
                    (1234567890123456789.0),
                    (0.1234567890123456789)
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb.query_df(decimals, "x", "select * from x").fetchall()
        assert conversion == reference
        print(reference)
        print(conversion)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_fallback_to_double(self, pandas, duckdb_cursor):
        # The widths of these decimal values are bigger than the max supported width for DECIMAL
        data = [
            Decimal("1.234567890123456789012345678901234567890123456789"),
            Decimal("123456789012345678901234567890123456789012345678.0"),
        ]
        decimals = pandas.DataFrame(data={"0": data})
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                    (1.234567890123456789012345678901234567890123456789),
                    (123456789012345678901234567890123456789012345678.0)
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb.query_df(decimals, "x", "select * from x").fetchall()
        assert conversion == reference
        assert isinstance(conversion[0][0], float)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_double_mixed(self, pandas, duckdb_cursor):
        data = [
            Decimal("1.234"),
            Decimal("1.234567891234567890123456789012345678901234567890123456789"),
            Decimal("0.00000000000345"),
            Decimal("0.00000000000000000000000000000000000000000000000000000000000123456789"),
            Decimal("1234543534535213412342342.2345456"),
            Decimal("123456789123456789123456789123456789123456789123456789123456789123456789"),
            Decimal("1232354.000000000000000000000000000035"),
            Decimal("123.5e300"),
        ]
        decimals = pandas.DataFrame(data={"0": data})
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                    (1.234),
                    (1.234567891234567890123456789012345678901234567890123456789),
                    (0.00000000000345),
                    (0.00000000000000000000000000000000000000000000000000000000000123456789),
                    (1234543534535213412342342.2345456),
                    (123456789123456789123456789123456789123456789123456789123456789123456789),
                    (1232354.000000000000000000000000000035),
                    (123.5e300)
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb_cursor.sql("select * from decimals").fetchall()
        assert conversion == reference
        assert isinstance(conversion[0][0], float)

    @pytest.mark.parametrize('pandas', [NumpyPandas(), ArrowPandas()])
    def test_numeric_decimal_out_of_range(self, pandas, duckdb_cursor):
        data = [Decimal("1.234567890123456789012345678901234567"), Decimal("123456789012345678901234567890123456.0")]
        decimals = pandas.DataFrame(data={"0": data})
        reference_query = """
            CREATE TABLE tbl AS SELECT * FROM (
                VALUES
                    (1.234567890123456789012345678901234567),
                    (123456789012345678901234567890123456.0)
            ) tbl(a);
        """
        duckdb_cursor.execute(reference_query)
        reference = duckdb_cursor.sql("select * from tbl").fetchall()
        conversion = duckdb_cursor.sql("select * from decimals").fetchall()
        assert conversion == reference



import duckdb


class TestNested(object):
    def test_lists(self, duckdb_cursor):
        result = duckdb_cursor.execute("SELECT LIST_VALUE(1, 2, 3, 4) ").fetchall()
        assert result == [([1, 2, 3, 4],)]

        result = duckdb_cursor.execute("SELECT LIST_VALUE() ").fetchall()
        assert result == [([],)]

        result = duckdb_cursor.execute("SELECT LIST_VALUE(1, 2, 3, NULL) ").fetchall()
        assert result == [([1, 2, 3, None],)]

    def test_nested_lists(self, duckdb_cursor):
        result = duckdb_cursor.execute("SELECT LIST_VALUE(LIST_VALUE(1, 2, 3, 4), LIST_VALUE(1, 2, 3, 4)) ").fetchall()
        assert result == [([[1, 2, 3, 4], [1, 2, 3, 4]],)]

        result = duckdb_cursor.execute(
            "SELECT LIST_VALUE(LIST_VALUE(1, 2, 3, 4), LIST_VALUE(1, 2, 3, NULL)) "
        ).fetchall()
        assert result == [([[1, 2, 3, 4], [1, 2, 3, None]],)]

    def test_struct(self, duckdb_cursor):
        result = duckdb_cursor.execute("SELECT STRUCT_PACK(a := 42, b := 43)").fetchall()
        assert result == [({'a': 42, 'b': 43},)]

        result = duckdb_cursor.execute("SELECT STRUCT_PACK(a := 42, b := NULL)").fetchall()
        assert result == [({'a': 42, 'b': None},)]

    def test_unnamed_struct(self, duckdb_cursor):
        result = duckdb_cursor.execute("SELECT row('aa','bb') AS x").fetchall()
        assert result == [(('aa', 'bb'),)]

        result = duckdb_cursor.execute("SELECT row('aa',NULL) AS x").fetchall()
        assert result == [(('aa', None),)]

    def test_nested_struct(self, duckdb_cursor):
        result = duckdb_cursor.execute("SELECT STRUCT_PACK(a := 42, b := LIST_VALUE(10, 9, 8, 7))").fetchall()
        assert result == [({'a': 42, 'b': [10, 9, 8, 7]},)]

        result = duckdb_cursor.execute("SELECT STRUCT_PACK(a := 42, b := LIST_VALUE(10, 9, 8, NULL))").fetchall()
        assert result == [({'a': 42, 'b': [10, 9, 8, None]},)]

    def test_map(self, duckdb_cursor):
        result = duckdb_cursor.execute("select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, 7))").fetchall()
        assert result == [({'key': [1, 2, 3, 4], 'value': [10, 9, 8, 7]},)]

        result = duckdb_cursor.execute("select MAP(LIST_VALUE(1, 2, 3, 4),LIST_VALUE(10, 9, 8, NULL))").fetchall()
        assert result == [({'key': [1, 2, 3, 4], 'value': [10, 9, 8, None]},)]

        result = duckdb_cursor.execute("SELECT MAP() ").fetchall()
        assert result == [({'key': [], 'value': []},)]



from typing import Optional, Union
from duckdb.experimental.spark._globals import _NoValueType, _NoValue
from duckdb import DuckDBPyConnection


class RuntimeConfig:
    def __init__(self, connection: DuckDBPyConnection):
        self._connection = connection

    def set(self, key: str, value: str) -> None:
        raise NotImplementedError

    def isModifiable(self, key: str) -> bool:
        raise NotImplementedError

    def unset(self, key: str) -> None:
        raise NotImplementedError

    def get(self, key: str, default: Union[Optional[str], _NoValueType] = _NoValue) -> str:
        raise NotImplementedError


__all__ = ["RuntimeConfig"]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
import duckdb
import re


class TestDataFrameDropDuplicates(object):
    def test_spark_drop_duplicates(self, spark):
        # Prepare Data
        data = [
            ("James", "Sales", 3000),
            ("Michael", "Sales", 4600),
            ("Robert", "Sales", 4100),
            ("Maria", "Finance", 3000),
            ("James", "Sales", 3000),
            ("Scott", "Finance", 3300),
            ("Jen", "Finance", 3900),
            ("Jeff", "Marketing", 3000),
            ("Kumar", "Marketing", 2000),
            ("Saif", "Sales", 4100),
        ]

        # Create DataFrame
        columns = ["employee_name", "department", "salary"]
        df = spark.createDataFrame(data=data, schema=columns)

        distinctDF = df.distinct().sort("employee_name")
        assert distinctDF.count() == 9
        res = distinctDF.collect()
        # James | Sales had a duplicate, has been removed
        expected = [
            Row(employee_name='James', department='Sales', salary=3000),
            Row(employee_name='Jeff', department='Marketing', salary=3000),
            Row(employee_name='Jen', department='Finance', salary=3900),
            Row(employee_name='Kumar', department='Marketing', salary=2000),
            Row(employee_name='Maria', department='Finance', salary=3000),
            Row(employee_name='Michael', department='Sales', salary=4600),
            Row(employee_name='Robert', department='Sales', salary=4100),
            Row(employee_name='Saif', department='Sales', salary=4100),
            Row(employee_name='Scott', department='Finance', salary=3300),
        ]
        assert res == expected

        df2 = df.dropDuplicates().sort("employee_name")
        assert df2.count() == 9
        res2 = df2.collect()
        assert res2 == res

        with pytest.raises(NotImplementedError):
            # DuckDBPyRelation does not have 'distinct_on' support yet
            dropDisDF = df.dropDuplicates(["department", "salary"])
            print("Distinct count of department & salary : " + str(dropDisDF.count()))
            res = dropDisDF.collect()
            print(res)



import duckdb
import os
import pytest

pd = pytest.importorskip("pandas")
pa = pytest.importorskip("pyarrow")
from typing import Union
import pyarrow.compute as pc
import uuid
import datetime
import numpy as np
import cmath

from duckdb.typing import *


class TestRemoveFunction(object):
    def test_not_created(self):
        con = duckdb.connect()
        with pytest.raises(
            duckdb.InvalidInputException,
            match="No function by the name of 'not_a_registered_function' was found in the list of registered functions",
        ):
            con.remove_function('not_a_registered_function')

    def test_double_remove(self):
        def func(x: int) -> int:
            return x

        con = duckdb.connect()
        con.create_function('func', func)
        con.sql('select func(42)')
        con.remove_function('func')
        with pytest.raises(
            duckdb.InvalidInputException,
            match="No function by the name of 'func' was found in the list of registered functions",
        ):
            con.remove_function('func')

        with pytest.raises(duckdb.CatalogException, match='Scalar Function with name func does not exist!'):
            con.sql('select func(42)')

    def test_use_after_remove(self):
        def func(x: int) -> int:
            return x

        con = duckdb.connect()
        con.create_function('func', func)
        rel = con.sql('select func(42)')
        con.remove_function('func')
        """
            Error: Catalog Error: Scalar Function with name func does not exist!
        """
        with pytest.raises(
            duckdb.InvalidInputException, match='Attempting to execute an unsuccessful or closed pending query result'
        ):
            res = rel.fetchall()

    def test_use_after_remove_and_recreation(self):
        def func(x: str) -> str:
            return x

        con = duckdb.connect()
        con.create_function('func', func)

        with pytest.raises(duckdb.BinderException, match='No function matches the given name'):
            rel1 = con.sql('select func(42)')
        rel2 = con.sql("select func('test'::VARCHAR)")
        con.remove_function('func')

        def also_func(x: int) -> int:
            return x

        con.create_function('func', also_func)
        with pytest.raises(duckdb.InvalidInputException, match='No function matches the given name'):
            res = rel2.fetchall()

    def test_overwrite_name(self):
        def func(x):
            return x

        con = duckdb.connect()
        # create first version of the function
        con.create_function('func', func, [BIGINT], BIGINT)

        # create relation that uses the function
        rel1 = con.sql("select func('3')")

        def other_func(x):
            return x

        with pytest.raises(
            duckdb.NotImplementedException,
            match="A function by the name of 'func' is already created, creating multiple functions with the same name is not supported yet, please remove it first",
        ):
            con.create_function('func', other_func, [VARCHAR], VARCHAR)

        con.remove_function('func')

        with pytest.raises(
            duckdb.InvalidInputException, match='Catalog Error: Scalar Function with name func does not exist!'
        ):
            # Attempted to execute the relation using the 'func' function, but it was deleted
            rel1.fetchall()

        con.create_function('func', other_func, [VARCHAR], VARCHAR)
        # create relation that uses the new version
        rel2 = con.sql("select func('test')")

        # execute both relations
        res1 = rel1.fetchall()
        res2 = rel2.fetchall()
        # This has been converted to string, because the previous version of the function no longer exists
        assert res1 == [('3',)]
        assert res2 == [('test',)]



import duckdb
import pytest


@pytest.fixture(autouse=True)
def setup_and_teardown_of_table(duckdb_cursor):
    duckdb_cursor.execute("create table win(id int, v int, t int, f float, s varchar);")
    duckdb_cursor.execute(
        """
        insert into win values
		(1, 1, 2, 0.54, 'h'),
		(1, 1, 1, 0.21, 'e'),
		(1, 2, 3, 0.001, 'l'),
		(2, 10, 4, 0.04, 'l'),
		(2, 11, -1, 10.45, 'o'),
		(3, -1, 0, 13.32, ','),
		(3, 5, -2, 9.87, 'wor'),
		(3, null, 10, 6.56, 'ld'); 
		"""
    )
    yield
    duckdb_cursor.execute("drop table win")


@pytest.fixture()
def table(duckdb_cursor):
    return duckdb_cursor.table("win")


class TestRAPIWindows:
    # general purpose win functions
    def test_row_number(self, table):
        result = table.row_number("over ()").execute().fetchall()
        expected = list(range(1, 9))
        assert len(result) == len(expected)
        assert all([r[0] == e for r, e in zip(result, expected)])
        result = table.row_number("over (partition by id order by t asc)", "id, v, t").order("id").execute().fetchall()
        expected = [
            (1, 1, 1, 1),
            (1, 1, 2, 2),
            (1, 2, 3, 3),
            (2, 11, -1, 1),
            (2, 10, 4, 2),
            (3, 5, -2, 1),
            (3, -1, 0, 2),
            (3, None, 10, 3),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_rank(self, table):
        result = table.rank("over ()").execute().fetchall()
        expected = [1] * 8
        assert len(result) == len(expected)
        assert all([r[0] == e for r, e in zip(result, expected)])
        result = table.rank("over (partition by id order by v asc)", "id, v").order("id").execute().fetchall()
        expected = [(1, 1, 1), (1, 1, 1), (1, 2, 3), (2, 10, 1), (2, 11, 2), (3, -1, 1), (3, 5, 2), (3, None, 3)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    @pytest.mark.parametrize("f", ["dense_rank", "rank_dense"])
    def test_dense_rank(self, table, f):
        result = getattr(table, f)("over ()").execute().fetchall()
        expected = [1] * 8
        assert len(result) == len(expected)
        assert all([r[0] == e for r, e in zip(result, expected)])
        result = getattr(table, f)("over (partition by id order by v asc)", "id, v").order("id").execute().fetchall()
        expected = [(1, 1, 1), (1, 1, 1), (1, 2, 2), (2, 10, 1), (2, 11, 2), (3, -1, 1), (3, 5, 2), (3, None, 3)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_percent_rank(self, table):
        result = table.percent_rank("over ()").execute().fetchall()
        expected = [0.0] * 8
        assert len(result) == len(expected)
        assert all([r[0] == e for r, e in zip(result, expected)])
        result = table.percent_rank("over (partition by id order by v asc)", "id, v").order("id").execute().fetchall()
        expected = [
            (1, 1, 0.0),
            (1, 1, 0.0),
            (1, 2, 1.0),
            (2, 10, 0.0),
            (2, 11, 1.0),
            (3, -1, 0.0),
            (3, 5, 0.5),
            (3, None, 1.0),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_cume_dist(self, table):
        result = table.cume_dist("over ()").execute().fetchall()
        expected = [1.0] * 8
        assert len(result) == len(expected)
        assert all([r[0] == e for r, e in zip(result, expected)])
        result = table.cume_dist("over (partition by id order by v asc)", "id, v").order("id").execute().fetchall()
        expected = [
            (1, 1, 2 / 3),
            (1, 1, 2 / 3),
            (1, 2, 1.0),
            (2, 10, 0.5),
            (2, 11, 1.0),
            (3, -1, 1 / 3),
            (3, 5, 2 / 3),
            (3, None, 1.0),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_ntile(self, table):
        result = table.n_tile("over (order by v)", 3, "v").execute().fetchall()
        expected = [(-1, 1), (1, 1), (1, 1), (2, 2), (5, 2), (10, 2), (11, 3), (None, 3)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_lag(self, table):
        result = (
            table.lag("v", "over (partition by id order by t asc)", projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, None),
            (1, 1, 2, 1),
            (1, 2, 3, 1),
            (2, 11, -1, None),
            (2, 10, 4, 11),
            (3, 5, -2, None),
            (3, -1, 0, 5),
            (3, None, 10, -1),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            table.lag("v", "over (partition by id order by t asc)", default_value="-1", projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, -1),
            (1, 1, 2, 1),
            (1, 2, 3, 1),
            (2, 11, -1, -1),
            (2, 10, 4, 11),
            (3, 5, -2, -1),
            (3, -1, 0, 5),
            (3, None, 10, -1),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            table.lag("v", "over (partition by id order by t asc)", offset=2, projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, None),
            (1, 1, 2, None),
            (1, 2, 3, 1),
            (2, 11, -1, None),
            (2, 10, 4, None),
            (3, 5, -2, None),
            (3, -1, 0, None),
            (3, None, 10, 5),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_lead(self, table):
        result = (
            table.lead("v", "over (partition by id order by t asc)", projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, 1),
            (1, 1, 2, 2),
            (1, 2, 3, None),
            (2, 11, -1, 10),
            (2, 10, 4, None),
            (3, 5, -2, -1),
            (3, -1, 0, None),
            (3, None, 10, None),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            table.lead("v", "over (partition by id order by t asc)", default_value="-1", projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, 1),
            (1, 1, 2, 2),
            (1, 2, 3, -1),
            (2, 11, -1, 10),
            (2, 10, 4, -1),
            (3, 5, -2, -1),
            (3, -1, 0, None),
            (3, None, 10, -1),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            table.lead("v", "over (partition by id order by t asc)", offset=2, projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, 2),
            (1, 1, 2, None),
            (1, 2, 3, None),
            (2, 11, -1, None),
            (2, 10, 4, None),
            (3, 5, -2, None),
            (3, -1, 0, None),
            (3, None, 10, None),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_first_value(self, table):
        result = (
            table.first_value("v", "over (partition by id order by t asc)", projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, 1),
            (1, 1, 2, 1),
            (1, 2, 3, 1),
            (2, 11, -1, 11),
            (2, 10, 4, 11),
            (3, 5, -2, 5),
            (3, -1, 0, 5),
            (3, None, 10, 5),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_last_value(self, table):
        result = (
            table.last_value(
                "v",
                "over (partition by id order by t asc range between unbounded preceding and unbounded following) ",
                projected_columns="id, v, t",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, 2),
            (1, 1, 2, 2),
            (1, 2, 3, 2),
            (2, 11, -1, 10),
            (2, 10, 4, 10),
            (3, 5, -2, None),
            (3, -1, 0, None),
            (3, None, 10, None),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_nth_value(self, table):
        result = (
            table.nth_value("v", "over (partition by id order by t asc)", offset=2, projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, None),
            (1, 1, 2, 1),
            (1, 2, 3, 1),
            (2, 11, -1, None),
            (2, 10, 4, 10),
            (3, 5, -2, None),
            (3, -1, 0, -1),
            (3, None, 10, -1),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            table.nth_value("v", "over (partition by id order by t asc)", offset=4, projected_columns="id, v, t")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, 1, 1, None),
            (1, 1, 2, None),
            (1, 2, 3, None),
            (2, 11, -1, None),
            (2, 10, 4, None),
            (3, 5, -2, None),
            (3, -1, 0, None),
            (3, None, 10, None),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    # agg functions within win
    def test_any_value(self, table):
        result = (
            table.any_value("v", window_spec="over (partition by id order by t asc)", projected_columns="id")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 1), (2, 11), (2, 11), (3, 5), (3, 5), (3, 5)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_arg_max(self, table):
        result = (
            table.arg_max("t", "v", window_spec="over (partition by id)", projected_columns="id")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 3), (1, 3), (1, 3), (2, -1), (2, -1), (3, -2), (3, -2), (3, -2)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_arg_min(self, table):
        result = (
            table.arg_min("t", "v", window_spec="over (partition by id)", projected_columns="id")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 2), (1, 2), (1, 2), (2, 4), (2, 4), (3, 0), (3, 0), (3, 0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_avg(self, table):
        result = [
            (r[0], round(r[1], 2))
            for r in (
                table.avg(
                    "v",
                    window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                    projected_columns="id",
                )
                .order("id")
                .execute()
                .fetchall()
            )
        ]
        expected = [(1, 1.0), (1, 1.0), (1, 1.33), (2, 11.0), (2, 10.5), (3, 5.0), (3, 2.0), (3, 2.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bit_and(self, table):
        result = (
            table.bit_and(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 0), (2, 11), (2, 10), (3, 5), (3, 5), (3, 5)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bit_or(self, table):
        result = (
            table.bit_or(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 3), (2, 11), (2, 11), (3, 5), (3, -1), (3, -1)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bit_xor(self, table):
        result = (
            table.bit_xor(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 0), (1, 2), (2, 11), (2, 1), (3, 5), (3, -6), (3, -6)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bitstring_agg(self, table):
        with pytest.raises(duckdb.BinderException, match="Could not retrieve required statistics"):
            result = (
                table.bitstring_agg(
                    "v",
                    window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                    projected_columns="id",
                )
                .order("id")
                .execute()
                .fetchall()
            )
        result = (
            table.bitstring_agg(
                "v",
                min=-1,
                max=11,
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, '0010000000000'),
            (1, '0010000000000'),
            (1, '0011000000000'),
            (2, '0000000000001'),
            (2, '0000000000011'),
            (3, '0000001000000'),
            (3, '1000001000000'),
            (3, '1000001000000'),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bool_and(self, table):
        result = (
            table.bool_and("t::BOOL", window_spec="over (partition by id)", projected_columns="id")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, True), (1, True), (1, True), (2, True), (2, True), (3, False), (3, False), (3, False)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_bool_or(self, table):
        result = (
            table.bool_or("t::BOOL", window_spec="over (partition by id)", projected_columns="id")
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, True), (1, True), (1, True), (2, True), (2, True), (3, True), (3, True), (3, True)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_count(self, table):
        result = (
            table.count(
                "id",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 2), (1, 3), (2, 1), (2, 2), (3, 1), (3, 2), (3, 3)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_favg(self, table):
        result = [
            (r[0], round(r[1], 2))
            for r in table.favg(
                "f",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, 0.21), (1, 0.38), (1, 0.25), (2, 10.45), (2, 5.24), (3, 9.87), (3, 11.59), (3, 9.92)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_fsum(self, table):
        result = [
            (r[0], round(r[1], 2))
            for r in table.fsum(
                "f",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, 0.21), (1, 0.75), (1, 0.75), (2, 10.45), (2, 10.49), (3, 9.87), (3, 23.19), (3, 29.75)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    @pytest.mark.skip(reason="geomean is not supported from a windowing context")
    def test_geomean(self, table):
        raise RuntimeError()

    def test_histogram(self, table):
        result = (
            table.histogram(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, {'key': [1], 'value': [1]}),
            (1, {'key': [1], 'value': [2]}),
            (1, {'key': [1, 2], 'value': [2, 1]}),
            (2, {'key': [11], 'value': [1]}),
            (2, {'key': [10, 11], 'value': [1, 1]}),
            (3, {'key': [5], 'value': [1]}),
            (3, {'key': [-1, 5], 'value': [1, 1]}),
            (3, {'key': [-1, 5], 'value': [1, 1]}),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_list(self, table):
        result = (
            table.list(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, [1]),
            (1, [1, 1]),
            (1, [1, 1, 2]),
            (2, [11]),
            (2, [11, 10]),
            (3, [5]),
            (3, [5, -1]),
            (3, [5, -1, None]),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_max(self, table):
        result = (
            table.max(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 2), (2, 11), (2, 11), (3, 5), (3, 5), (3, 5)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_min(self, table):
        result = (
            table.min(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 1), (2, 11), (2, 10), (3, 5), (3, -1), (3, -1)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_product(self, table):
        result = (
            table.product(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 1), (1, 2), (2, 11), (2, 110), (3, 5), (3, -5), (3, -5)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_string_agg(self, table):
        result = (
            table.string_agg(
                "s",
                sep="/",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 'e'), (1, 'e/h'), (1, 'e/h/l'), (2, 'o'), (2, 'o/l'), (3, 'wor'), (3, 'wor/,'), (3, 'wor/,/ld')]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_sum(self, table):
        result = (
            table.sum(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1), (1, 2), (1, 4), (2, 11), (2, 21), (3, 5), (3, 4), (3, 4)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_median(self, table):
        result = (
            table.median(
                "v",
                window_spec="over (partition by id order by t asc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 1.0), (1, 1.0), (1, 1.0), (2, 11.0), (2, 10.5), (3, 5.0), (3, 2.0), (3, 2.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_mode(self, table):
        result = (
            table.mode(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 2), (1, 2), (1, 1), (2, 10), (2, 10), (3, None), (3, -1), (3, -1)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_quantile_cont(self, table):
        result = (
            table.quantile_cont(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 2.0), (1, 1.5), (1, 1.0), (2, 10.0), (2, 10.5), (3, None), (3, -1.0), (3, 2.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = [
            (r[0], list(map(lambda x: round(x, 2), r[1])) if r[1] is not None else None)
            for r in table.quantile_cont(
                "v",
                q=[0.2, 0.5],
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [
            (1, [2.0, 2.0]),
            (1, [1.2, 1.5]),
            (1, [1.0, 1.0]),
            (2, [10.0, 10.0]),
            (2, [10.2, 10.5]),
            (3, None),
            (3, [-1.0, -1.0]),
            (3, [0.2, 2.0]),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    @pytest.mark.parametrize("f", ["quantile_disc", "quantile"])
    def test_quantile_disc(self, table, f):
        result = (
            getattr(table, f)(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [(1, 2), (1, 1), (1, 1), (2, 10), (2, 10), (3, None), (3, -1), (3, -1)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])
        result = (
            getattr(table, f)(
                "v",
                q=[0.2, 0.8],
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        )
        expected = [
            (1, [2, 2]),
            (1, [1, 2]),
            (1, [1, 2]),
            (2, [10, 10]),
            (2, [10, 11]),
            (3, None),
            (3, [-1, -1]),
            (3, [-1, 5]),
        ]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_stddev_pop(self, table):
        result = [
            (r[0], round(r[1], 2)) if r[1] is not None else r
            for r in table.stddev_pop(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, 0.0), (1, 0.5), (1, 0.47), (2, 0.0), (2, 0.5), (3, None), (3, 0.0), (3, 3.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    @pytest.mark.parametrize("f", ["stddev_samp", "stddev", "std"])
    def test_stddev_samp(self, table, f):
        result = [
            (r[0], round(r[1], 2)) if r[1] is not None else r
            for r in getattr(table, f)(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, None), (1, 0.71), (1, 0.58), (2, None), (2, 0.71), (3, None), (3, None), (3, 4.24)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    def test_var_pop(self, table):
        result = [
            (r[0], round(r[1], 2)) if r[1] is not None else r
            for r in table.var_pop(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, 0.0), (1, 0.25), (1, 0.22), (2, 0.0), (2, 0.25), (3, None), (3, 0.0), (3, 9.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])

    @pytest.mark.parametrize("f", ["var_samp", "variance", "var"])
    def test_var_samp(self, table, f):
        result = [
            (r[0], round(r[1], 2)) if r[1] is not None else r
            for r in getattr(table, f)(
                "v",
                window_spec="over (partition by id order by t desc rows between unbounded preceding and current row)",
                projected_columns="id",
            )
            .order("id")
            .execute()
            .fetchall()
        ]
        expected = [(1, None), (1, 0.5), (1, 0.33), (2, None), (2, 0.5), (3, None), (3, None), (3, 18.0)]
        assert len(result) == len(expected)
        assert all([r == e for r, e in zip(result, expected)])



import duckdb
import os

try:
    import pyarrow
    import pyarrow.parquet
    import numpy as np

    can_run = True
except:
    can_run = False


class TestArrowParallel(object):
    def test_parallel_run(self, duckdb_cursor):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")
        data = pyarrow.array(np.random.randint(800, size=1000000), type=pyarrow.int32())
        tbl = pyarrow.Table.from_batches(pyarrow.Table.from_arrays([data], ['a']).to_batches(10000))
        rel = duckdb_conn.from_arrow(tbl)
        # Also test multiple reads
        assert rel.aggregate("(count(a))::INT").execute().fetchone()[0] == 1000000
        assert rel.aggregate("(count(a))::INT").execute().fetchone()[0] == 1000000

    def test_parallel_types_and_different_batches(self, duckdb_cursor):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')
        cols = 'id, first_name, last_name, email, gender, ip_address, cc, country, birthdate, salary, title, comments'

        userdata_parquet_table = pyarrow.parquet.read_table(parquet_filename)
        for i in [7, 51, 99, 100, 101, 500, 1000, 2000]:
            data = pyarrow.array(np.arange(3, 7), type=pyarrow.int32())
            tbl = pyarrow.Table.from_arrays([data], ['a'])
            rel_id = duckdb_conn.from_arrow(tbl)
            userdata_parquet_table2 = pyarrow.Table.from_batches(userdata_parquet_table.to_batches(i))
            rel = duckdb_conn.from_arrow(userdata_parquet_table2)
            result = rel.filter("first_name=\'Jose\' and salary > 134708.82").aggregate('count(*)')
            assert result.execute().fetchone()[0] == 4

    def test_parallel_fewer_batches_than_threads(self, duckdb_cursor):
        if not can_run:
            return
        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")
        duckdb_conn.execute("PRAGMA verify_parallelism")

        data = pyarrow.array(np.random.randint(800, size=1000), type=pyarrow.int32())
        tbl = pyarrow.Table.from_batches(pyarrow.Table.from_arrays([data], ['a']).to_batches(2))
        rel = duckdb_conn.from_arrow(tbl)
        # Also test multiple reads
        assert rel.aggregate("(count(a))::INT").execute().fetchone()[0] == 1000



import pandas as pd
import numpy as np
import duckdb
import os
import tempfile


class TestNonDefaultConn(object):
    def test_values(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb.values([1], duckdb_cursor).insert_into("t")
        assert duckdb_cursor.execute("select count(*) from t").fetchall()[0] == (1,)

    def test_query(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        assert duckdb_cursor.query("select count(*) from t").execute().fetchall()[0] == (1,)
        assert duckdb_cursor.from_query("select count(*) from t").execute().fetchall()[0] == (1,)

    def test_from_csv(self, duckdb_cursor):
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        test_df.to_csv(temp_file_name, index=False)
        rel = duckdb_cursor.from_csv_auto(temp_file_name)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)

    def test_from_parquet(self, duckdb_cursor):
        try:
            import pyarrow as pa
        except ImportError:
            return
        temp_file_name = os.path.join(tempfile.mkdtemp(), next(tempfile._get_candidate_names()))
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        test_df.to_parquet(temp_file_name, index=False)
        rel = duckdb_cursor.from_parquet(temp_file_name)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)

    def test_from_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        rel = duckdb.df(test_df, connection=duckdb_cursor)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)
        rel = duckdb_cursor.from_df(test_df)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)

    def test_from_arrow(self, duckdb_cursor):
        try:
            import pyarrow as pa
        except:
            return

        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        test_arrow = pa.Table.from_pandas(test_df)
        rel = duckdb_cursor.from_arrow(test_arrow)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)
        rel = duckdb.arrow(test_arrow, connection=duckdb_cursor)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)

    def test_filter_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1), (4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        rel = duckdb.filter(test_df, "i < 2", connection=duckdb_cursor)
        assert rel.query('t_2', 'select count(*) from t inner join t_2 on (a = i)').fetchall()[0] == (1,)

    def test_project_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1), (4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4], "j": [1, 2, 3, 4]})
        rel = duckdb.project(test_df, "i", connection=duckdb_cursor)
        assert rel.query('t_2', 'select * from t inner join t_2 on (a = i)').fetchall()[0] == (1, 1)

    def test_agg_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1), (4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4], "j": [1, 2, 3, 4]})
        rel = duckdb.aggregate(test_df, "count(*) as i", connection=duckdb_cursor)
        assert rel.query('t_2', 'select * from t inner join t_2 on (a = i)').fetchall()[0] == (4, 4)

    def test_distinct_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1)")
        test_df = pd.DataFrame.from_dict({"i": [1, 1, 2, 3, 4]})
        rel = duckdb.distinct(test_df, connection=duckdb_cursor)
        assert rel.query('t_2', 'select * from t inner join t_2 on (a = i)').fetchall()[0] == (1, 1)

    def test_limit_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1),(4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        rel = duckdb.limit(test_df, 1, connection=duckdb_cursor)
        assert rel.query('t_2', 'select * from t inner join t_2 on (a = i)').fetchall()[0] == (1, 1)

    def test_query_df(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1),(4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        rel = duckdb.query_df(test_df, 't_2', 'select * from t inner join t_2 on (a = i)', connection=duckdb_cursor)
        assert rel.fetchall()[0] == (1, 1)

    def test_query_order(self, duckdb_cursor):
        duckdb_cursor.execute("create table t (a integer)")
        duckdb_cursor.execute("insert into t values (1),(4)")
        test_df = pd.DataFrame.from_dict({"i": [1, 2, 3, 4]})
        rel = duckdb.order(test_df, 'i', connection=duckdb_cursor)
        assert rel.query('t_2', 'select * from t inner join t_2 on (a = i)').fetchall()[0] == (1, 1)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit, array_contains
from duckdb.experimental.spark.sql.functions import sum, avg, max, min, mean, count


@pytest.fixture
def array_df(spark):
    data = [
        ("James,,Smith", ["Java", "Scala", "C++"], ["Spark", "Java"]),
        ("Michael,Rose,", ["Spark", "Java", "C++"], ["Spark", "Java"]),
        ("Robert,,Williams", ["CSharp", "VB"], ["Spark", "Python"]),
    ]
    dataframe = spark.createDataFrame(data=data, schema=["Name", "Languages1", "Languages2"])
    yield dataframe


@pytest.fixture
def df(spark):
    simpleData = (
        ("Java", 4000, 5),
        ("Python", 4600, 10),
        ("Scala", 4100, 15),
        ("Scala", 4500, 15),
        ("PHP", 3000, 20),
    )
    columns = ["CourseName", "fee", "discount"]
    dataframe = spark.createDataFrame(data=simpleData, schema=columns)
    yield dataframe


class TestDataFrameUnion(object):
    def test_transform(self, spark, df):
        # Custom transformation 1
        from duckdb.experimental.spark.sql.functions import upper

        def to_upper_str_columns(df):
            return df.withColumn("CourseName", upper(df.CourseName))

        # Custom transformation 2
        def reduce_price(df, reduceBy):
            return df.withColumn("new_fee", df.fee - reduceBy)

        # Custom transformation 3
        def apply_discount(df):
            return df.withColumn("discounted_fee", df.new_fee - (df.new_fee * df.discount) / 100)

        df2 = df.transform(to_upper_str_columns).transform(reduce_price, 1000).transform(apply_discount)
        res = df2.collect()
        assert res == [
            Row(CourseName='JAVA', fee=4000, discount=5, new_fee=3000, discounted_fee=2850.0),
            Row(CourseName='PYTHON', fee=4600, discount=10, new_fee=3600, discounted_fee=3240.0),
            Row(CourseName='SCALA', fee=4100, discount=15, new_fee=3100, discounted_fee=2635.0),
            Row(CourseName='SCALA', fee=4500, discount=15, new_fee=3500, discounted_fee=2975.0),
            Row(CourseName='PHP', fee=3000, discount=20, new_fee=2000, discounted_fee=1600.0),
        ]

    # https://sparkbyexamples.com/pyspark/pyspark-transform-function/
    @pytest.mark.skip(reason='LambdaExpressions are currently under development, waiting til that is finished')
    def test_transform_function(self, spark, array_df):
        from duckdb.experimental.spark.sql.functions import upper, transform

        df.select(transform("Languages1", lambda x: upper(x)).alias("languages1")).show()



import numpy as np
import datetime
import duckdb
import pytest

pandas = pytest.importorskip("pandas")


class TestPandasNaN(object):
    def test_pandas_nan(self, duckdb_cursor):
        # create a DataFrame with some basic values
        df = pandas.DataFrame([{"col1": "val1", "col2": 1.05}, {"col1": "val3", "col2": np.NaN}])
        # create a new column (newcol1) that includes either NaN or values from col1
        df["newcol1"] = np.where(df["col1"] == "val1", np.NaN, df["col1"])
        # now create a new column with the current time
        # (FIXME: we replace the microseconds with 0 for now, because we only support milisecond resolution)
        current_time = datetime.datetime.now().replace(microsecond=0)
        df['datetest'] = current_time
        # introduce a NaT (Not a Time value)
        df.loc[0, 'datetest'] = pandas.NaT
        # now pass the DF through duckdb:

        conn = duckdb.connect(':memory:')
        conn.register('testing_null_values', df)
        # scan the DF and fetch the results normally
        results = conn.execute('select * from testing_null_values').fetchall()
        assert results[0][0] == 'val1'
        assert results[0][1] == 1.05
        assert results[0][2] == None
        assert results[0][3] == None
        assert results[1][0] == 'val3'
        assert results[1][1] == None
        assert results[1][2] == 'val3'
        assert results[1][3] == current_time

        # now fetch the results as numpy:
        result_np = conn.execute('select * from testing_null_values').fetchnumpy()
        assert result_np['col1'][0] == df['col1'][0]
        assert result_np['col1'][1] == df['col1'][1]
        assert result_np['col2'][0] == df['col2'][0]

        assert result_np['col2'].mask[1]
        assert result_np['newcol1'].mask[0]
        assert result_np['newcol1'][1] == df['newcol1'][1]

        result_df = conn.execute('select * from testing_null_values').fetchdf()
        assert pandas.isnull(result_df['datetest'][0])
        assert result_df['datetest'][1] == df['datetest'][1]



import duckdb
import os

try:
    import pyarrow as pa
    from pyarrow import parquet as pq
    import numpy as np

    can_run = True
except:
    can_run = False


def create_binary_table(type):
    schema = pa.schema([("data", type)])
    inputs = [pa.array([b"foo", b"bar", b"baz"], type=type)]
    return pa.Table.from_arrays(inputs, schema=schema)


class TestArrowBinary(object):
    def test_binary_types(self, duckdb_cursor):
        if not can_run:
            return

        # Fixed Size Binary
        arrow_table = create_binary_table(pa.binary(3))
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [(b"foo",), (b"bar",), (b"baz",)]

        # Normal Binary
        arrow_table = create_binary_table(pa.binary())
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [(b"foo",), (b"bar",), (b"baz",)]

        # Large Binary
        arrow_table = create_binary_table(pa.large_binary())
        rel = duckdb.from_arrow(arrow_table)
        res = rel.execute().fetchall()
        assert res == [(b"foo",), (b"bar",), (b"baz",)]



import duckdb
from pandas import DataFrame
import pytest


class TestInsertInto(object):
    def test_insert_into_schema(self, duckdb_cursor):
        # open connection
        con = duckdb.connect()
        con.execute('CREATE SCHEMA s')
        con.execute('CREATE TABLE s.t (id INTEGER PRIMARY KEY)')

        # make relation
        df = DataFrame([1], columns=['id'])
        rel = con.from_df(df)

        rel.insert_into('s.t')

        assert con.execute("select * from s.t").fetchall() == [(1,)]

        # This should fail since this will go to default schema
        with pytest.raises(duckdb.CatalogException):
            rel.insert_into('t')

        # If we add t in the default schema it should work.
        con.execute('CREATE TABLE t (id INTEGER PRIMARY KEY)')
        rel.insert_into('t')
        assert con.execute("select * from t").fetchall() == [(1,)]



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")
from duckdb.experimental.spark.sql.types import Row


class TestReplaceValue(object):
    # https://sparkbyexamples.com/pyspark/pyspark-replace-column-values/?expand_article=1
    def test_replace_value(self, spark):
        address = [(1, "14851 Jeffrey Rd", "DE"), (2, "43421 Margarita St", "NY"), (3, "13111 Siemon Ave", "CA")]
        df = spark.createDataFrame(address, ["id", "address", "state"])

        # Replace part of string with another string
        from duckdb.experimental.spark.sql.functions import regexp_replace

        df2 = df.withColumn('address', regexp_replace('address', 'Rd', 'Road'))

        # Replace string column value conditionally
        from duckdb.experimental.spark.sql.functions import when

        res = df2.collect()
        print(res)
        df2 = df.withColumn(
            'address',
            when(df.address.endswith('Rd'), regexp_replace(df.address, 'Rd', 'Road'))
            .when(df.address.endswith('St'), regexp_replace(df.address, 'St', 'Street'))
            .when(df.address.endswith('Ave'), regexp_replace(df.address, 'Ave', 'Avenue'))
            .otherwise(df.address),
        )
        res = df2.collect()
        print(res)
        expected = [
            Row(id=1, address='14851 Jeffrey Road', state='DE'),
            Row(id=2, address='43421 Margarita Street', state='NY'),
            Row(id=3, address='13111 Siemon Avenue', state='CA'),
        ]
        print(expected)
        assert res == expected
        # Replace all substrings of the specified string value that match regexp with rep.
        df3 = spark.createDataFrame([('100-200',)], ['str'])
        res = df3.select(regexp_replace('str', r'(\d+)', '--').alias('d')).collect()
        expected = [Row(d='-----')]
        print(expected)
        assert res == expected



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    Row,
)


class TestDataFrameLimit(object):
    def test_dataframe_limit(self, spark):
        df = spark.sql("select * from range(100000)")
        df2 = df.limit(10)
        res = df2.collect()
        expected = [
            Row(range=0),
            Row(range=1),
            Row(range=2),
            Row(range=3),
            Row(range=4),
            Row(range=5),
            Row(range=6),
            Row(range=7),
            Row(range=8),
            Row(range=9),
        ]
        assert res == expected



import pandas as pd
import pytest
import duckdb


class TestPandasEnum(object):
    def test_3480(self, duckdb_cursor):
        duckdb_cursor.execute(
            """
        create type cat as enum ('marie', 'duchess', 'toulouse');
        create table tab (
            cat cat,
            amt int
        );
        """
        )
        df = duckdb_cursor.query(f"SELECT * FROM tab LIMIT 0;").to_df()
        assert df["cat"].cat.categories.equals(pd.Index(['marie', 'duchess', 'toulouse']))
        duckdb_cursor.execute("DROP TABLE tab")
        duckdb_cursor.execute("DROP TYPE cat")

    def test_3479(self, duckdb_cursor):
        duckdb_cursor.execute(
            """
        create type cat as enum ('marie', 'duchess', 'toulouse');
        create table tab (
            cat cat,
            amt int
        );
        """
        )

        df = pd.DataFrame(
            {
                "cat2": pd.Series(['duchess', 'toulouse', 'marie', None, "berlioz", "o_malley"], dtype="category"),
                "amt": [1, 2, 3, 4, 5, 6],
            }
        )
        duckdb_cursor.register('df', df)
        with pytest.raises(
            duckdb.ConversionException,
            match='Type UINT8 with value 0 can\'t be cast because the value is out of range for the destination type UINT8',
        ):
            duckdb_cursor.execute(f"INSERT INTO tab SELECT * FROM df;")

        assert duckdb_cursor.execute("select * from tab").fetchall() == []
        duckdb_cursor.execute("DROP TABLE tab")
        duckdb_cursor.execute("DROP TYPE cat")



from typing import Optional, List, Tuple, Any, Union, Iterable, TYPE_CHECKING
import uuid

if TYPE_CHECKING:
    from .catalog import Catalog
    from pandas.core.frame import DataFrame as PandasDataFrame

from ..exception import ContributionsAcceptedError
 
from .types import StructType, AtomicType, DataType
from ..conf import SparkConf
from .dataframe import DataFrame
from .conf import RuntimeConfig
from .readwriter import DataFrameReader
from ..context import SparkContext
from .udf import UDFRegistration
from .streaming import DataStreamReader
import duckdb

from ..errors import (
    PySparkTypeError,
    PySparkValueError
)

from ..errors.error_classes import *

# In spark:
# SparkSession holds a SparkContext
# SparkContext gets created from SparkConf
# At this level the check is made to determine whether the instance already exists and just needs to be retrieved or it needs to be created

# For us this is done inside of `duckdb.connect`, based on the passed in path + configuration
# SparkContext can be compared to our Connection class, and SparkConf to our ClientContext class


# data is a List of rows
# every value in each row needs to be turned into a Value
def _combine_data_and_schema(data: Iterable[Any], schema: StructType):
    from duckdb import Value

    new_data = []
    for row in data:
        new_row = [Value(x, dtype.duckdb_type) for x, dtype in zip(row, [y.dataType for y in schema])]
        new_data.append(new_row)
    return new_data


class SparkSession:
    def __init__(self, context: SparkContext):
        self.conn = context.connection
        self._context = context
        self._conf = RuntimeConfig(self.conn)

    def _create_dataframe(self, data: Union[Iterable[Any], "PandasDataFrame"]) -> DataFrame:
        try:
            import pandas

            has_pandas = True
        except:
            has_pandas = False
        if has_pandas and isinstance(data, pandas.DataFrame):
            unique_name = f'pyspark_pandas_df_{uuid.uuid1()}'
            self.conn.register(unique_name, data)
            return DataFrame(self.conn.sql(f'select * from "{unique_name}"'), self)

        def verify_tuple_integrity(tuples):
            if len(tuples) <= 1:
                return
            expected_length = len(tuples[0])
            for i, item in enumerate(tuples[1:]):
                actual_length = len(item)
                if expected_length == actual_length:
                    continue
                raise PySparkTypeError(
                    error_class="LENGTH_SHOULD_BE_THE_SAME",
                    message_parameters={
                        "arg1": f"data{i}",
                        "arg2": f"data{i+1}",
                        "arg1_length": str(expected_length),
                        "arg2_length": str(actual_length)
                    },
                )

        if not isinstance(data, list):
            data = list(data)
        verify_tuple_integrity(data)

        def construct_query(tuples) -> str:
            def construct_values_list(row, start_param_idx):
                parameter_count = len(row)
                parameters = [f'${x+start_param_idx}' for x in range(parameter_count)]
                parameters = '(' + ', '.join(parameters) + ')'
                return parameters

            row_size = len(tuples[0])
            values_list = [construct_values_list(x, 1 + (i * row_size)) for i, x in enumerate(tuples)]
            values_list = ', '.join(values_list)

            query = f"""
                select * from (values {values_list})
            """
            return query

        query = construct_query(data)

        def construct_parameters(tuples):
            parameters = []
            for row in tuples:
                parameters.extend(list(row))
            return parameters

        parameters = construct_parameters(data)

        rel = self.conn.sql(query, params=parameters)
        return DataFrame(rel, self)

    def _createDataFrameFromPandas(self, data: "PandasDataFrame", types, names) -> DataFrame:
        df = self._create_dataframe(data)

        # Cast to types
        if types:
            df = df._cast_types(*types)
        # Alias to names
        if names:
            df = df.toDF(*names)
        return df

    def createDataFrame(
        self,
        data: Union["PandasDataFrame", Iterable[Any]],
        schema: Optional[Union[StructType, List[str]]] = None,
        samplingRatio: Optional[float] = None,
        verifySchema: bool = True,
    ) -> DataFrame:
        if samplingRatio:
            raise NotImplementedError
        if not verifySchema:
            raise NotImplementedError
        types = None
        names = None

        if isinstance(data, DataFrame):
            raise PySparkTypeError(
                error_class="SHOULD_NOT_DATAFRAME",
                message_parameters={"arg_name": "data"},
            )

        if schema:
            if isinstance(schema, StructType):
                types, names = schema.extract_types_and_names()
            else:
                names = schema

        try:
            import pandas

            has_pandas = True
        except:
            has_pandas = False
        # Falsey check on pandas dataframe is not defined, so first check if it's not a pandas dataframe
        # Then check if 'data' is None or []
        if has_pandas and isinstance(data, pandas.DataFrame):
            return self._createDataFrameFromPandas(data, types, names)

        # Finally check if a schema was provided
        is_empty = False
        if not data and names:
            # Create NULLs for every type in our dataframe
            is_empty = True
            data = [tuple(None for _ in names)]

        if schema and isinstance(schema, StructType):
            # Transform the data into Values to combine the data+schema
            data = _combine_data_and_schema(data, schema)

        df = self._create_dataframe(data)
        if is_empty:
            rel = df.relation
            # Add impossible where clause
            rel = rel.filter('1=0')
            df = DataFrame(rel, self)

        # Cast to types
        if types:
            df = df._cast_types(*types)
        # Alias to names
        if names:
            df = df.toDF(*names)
        return df

    def newSession(self) -> "SparkSession":
        return SparkSession(self._context)

    def range(
        self, start: int, end: Optional[int] = None, step: int = 1, numPartitions: Optional[int] = None
    ) -> "DataFrame":
        raise ContributionsAcceptedError

    def sql(self, sqlQuery: str, **kwargs: Any) -> DataFrame:
        if kwargs:
            raise NotImplementedError
        relation = self.conn.sql(sqlQuery)
        return DataFrame(relation, self)

    def stop(self) -> None:
        self._context.stop()

    def table(self, tableName: str) -> DataFrame:
        relation = self.conn.table(tableName)
        return DataFrame(relation, self)

    def getActiveSession(self) -> "SparkSession":
        return self

    @property
    def catalog(self) -> "Catalog":
        if not hasattr(self, "_catalog"):
            from duckdb.experimental.spark.sql.catalog import Catalog

            self._catalog = Catalog(self)
        return self._catalog

    @property
    def conf(self) -> RuntimeConfig:
        return self._conf

    @property
    def read(self) -> DataFrameReader:
        return DataFrameReader(self)

    @property
    def readStream(self) -> DataStreamReader:
        return DataStreamReader(self)

    @property
    def sparkContext(self) -> SparkContext:
        return self._context

    @property
    def streams(self) -> Any:
        raise ContributionsAcceptedError

    @property
    def udf(self) -> UDFRegistration:
        return UDFRegistration()

    @property
    def version(self) -> str:
        return '1.0.0'

    class Builder:
        def __init__(self):
            pass

        def master(self, name: str) -> "SparkSession.Builder":
            # no-op
            return self

        def appName(self, name: str) -> "SparkSession.Builder":
            # no-op
            return self

        def remote(self, url: str) -> "SparkSession.Builder":
            # no-op
            return self

        def getOrCreate(self) -> "SparkSession":
            context = SparkContext("__ignored__")
            return SparkSession(context)

        def config(
            self, key: Optional[str] = None, value: Optional[Any] = None, conf: Optional[SparkConf] = None
        ) -> "SparkSession.Builder":
            return self

        def enableHiveSupport(self) -> "SparkSession.Builder":
            # no-op
            return self

    builder = Builder()


__all__ = ["SparkSession"]



import duckdb
import numpy


class TestBlob(object):
    def test_blob(self, duckdb_cursor):
        duckdb_cursor.execute("SELECT BLOB 'hello'")
        results = duckdb_cursor.fetchall()
        assert results[0][0] == b'hello'

        duckdb_cursor.execute("SELECT BLOB 'hello' AS a")
        results = duckdb_cursor.fetchnumpy()
        assert results['a'] == numpy.array([b'hello'], dtype=object)



import duckdb
import pytest

pq = pytest.importorskip("pyarrow.parquet")
pa = pytest.importorskip("pyarrow")

from datetime import time
from pathlib import PurePosixPath


class Test9443(object):
    def test_9443(self, tmp_path, duckdb_cursor):
        arrow_table = pa.Table.from_pylist(
            [
                {"col1": time(1, 2, 3)},
            ]
        )  # col1: time64[us]

        print(arrow_table)

        temp_file = str(PurePosixPath(tmp_path.as_posix()) / "test9443.parquet")
        pq.write_table(arrow_table, temp_file)

        sql = f'SELECT * FROM "{temp_file}"'

        duckdb_cursor.execute(sql)
        with pytest.raises(Exception, match='Unsupported Arrow type TIME WITH TIME ZONE'):
            duckdb_cursor.fetch_record_batch()



import duckdb
import os

try:
    import pyarrow
    import pyarrow.parquet
    import pyarrow.dataset
    from pyarrow.dataset import Scanner
    import pyarrow.compute as pc
    import numpy as np

    can_run = True
except:
    can_run = False


class TestArrowScanner(object):
    def test_parallel_scanner(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        arrow_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        scanner_filter = (pc.field("first_name") == pc.scalar('Jose')) & (pc.field("salary") > pc.scalar(134708.82))

        arrow_scanner = Scanner.from_dataset(arrow_dataset, filter=scanner_filter)

        rel = duckdb_conn.from_arrow(arrow_scanner)

        assert rel.aggregate('count(*)').execute().fetchone()[0] == 12

    def test_parallel_scanner_replacement_scans(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        arrow_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        scanner_filter = (pc.field("first_name") == pc.scalar('Jose')) & (pc.field("salary") > pc.scalar(134708.82))

        arrow_scanner = Scanner.from_dataset(arrow_dataset, filter=scanner_filter)

        assert duckdb_conn.execute("select count(*) from arrow_scanner").fetchone()[0] == 12

    def test_parallel_scanner_register(self, duckdb_cursor):
        if not can_run:
            return

        duckdb_conn = duckdb.connect()
        duckdb_conn.execute("PRAGMA threads=4")

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        arrow_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        scanner_filter = (pc.field("first_name") == pc.scalar('Jose')) & (pc.field("salary") > pc.scalar(134708.82))

        arrow_scanner = Scanner.from_dataset(arrow_dataset, filter=scanner_filter)

        duckdb_conn.register("bla", arrow_scanner)

        assert duckdb_conn.execute("select count(*) from bla").fetchone()[0] == 12

    def test_parallel_scanner_default_conn(self, duckdb_cursor):
        if not can_run:
            return

        parquet_filename = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'userdata1.parquet')

        arrow_dataset = pyarrow.dataset.dataset(
            [
                parquet_filename,
                parquet_filename,
                parquet_filename,
            ],
            format="parquet",
        )

        scanner_filter = (pc.field("first_name") == pc.scalar('Jose')) & (pc.field("salary") > pc.scalar(134708.82))

        arrow_scanner = Scanner.from_dataset(arrow_dataset, filter=scanner_filter)

        rel = duckdb.from_arrow(arrow_scanner)

        assert rel.aggregate('count(*)').execute().fetchone()[0] == 12



import duckdb
import os
import pandas as pd
import pytest

from duckdb.typing import *


class TestNativeUDF(object):
    def test_default_conn(self):
        def passthrough(x):
            return x

        duckdb.create_function('default_conn_passthrough', passthrough, [BIGINT], BIGINT)
        res = duckdb.sql('select default_conn_passthrough(5)').fetchall()
        assert res == [(5,)]

    def test_basic_use(self):
        def plus_one(x):
            if x == None or x > 50:
                return x
            return x + 1

        con = duckdb.connect()
        con.create_function('plus_one', plus_one, [BIGINT], BIGINT)
        assert [(6,)] == con.sql('select plus_one(5)').fetchall()

        range_table = con.table_function('range', [5000])
        res = con.sql('select plus_one(i) from range_table tbl(i)').fetchall()
        assert len(res) == 5000

        vector_size = duckdb.__standard_vector_size__
        res = con.sql(f'select i, plus_one(i) from test_vector_types(NULL::BIGINT, false) t(i), range({vector_size})')
        assert len(res) == (vector_size * 11)

    def test_passthrough(self):
        def passthrough(x):
            return x

        con = duckdb.connect()
        con.create_function('passthrough', passthrough, [BIGINT], BIGINT)
        assert (
            con.sql('select passthrough(i) from range(5000) tbl(i)').fetchall()
            == con.sql('select * from range(5000)').fetchall()
        )

    def test_execute(self):
        def func(x):
            return x % 2

        con = duckdb.connect()
        con.create_function('modulo_op', func, [BIGINT], TINYINT)
        res = con.execute('select modulo_op(?)', [5]).fetchall()
        assert res == [(1,)]

    def test_cast_output(self):
        def takes_string(x):
            return x

        con = duckdb.connect()
        con.create_function('casts_from_string', takes_string, [VARCHAR], BIGINT)

        res = con.sql("select casts_from_string('42')").fetchall()
        assert res == [(42,)]

        with pytest.raises(duckdb.InvalidInputException):
            res = con.sql("select casts_from_string('test')").fetchall()

    def test_detected_parameters(self):
        def concatenate(a: str, b: str):
            return a + b

        con = duckdb.connect()
        con.create_function('py_concatenate', concatenate, None, VARCHAR)
        res = con.sql(
            """
            select py_concatenate('5','3');
        """
        ).fetchall()
        assert res[0][0] == '53'

    def test_detected_return_type(self):
        def add_nums(*args) -> int:
            sum = 0
            for arg in args:
                sum += arg
            return sum

        con = duckdb.connect()
        con.create_function('add_nums', add_nums)
        res = con.sql(
            """
            select add_nums(5,3,2,1);
        """
        ).fetchall()
        assert res[0][0] == 11

    def test_varargs(self):
        def variable_args(*args):
            amount = len(args)
            return amount

        con = duckdb.connect()
        con.create_function('varargs', variable_args, None, BIGINT)
        res = con.sql("""select varargs('5', '3', '2', 1, 0.12345)""").fetchall()
        assert res == [(5,)]

    def test_return_incorrectly_typed_object(self):
        def returns_duckdb() -> int:
            return 'duckdb'

        con = duckdb.connect()
        con.create_function('fastest_database_in_the_west', returns_duckdb)
        with pytest.raises(
            duckdb.InvalidInputException, match="Failed to cast value: Could not convert string 'duckdb' to INT64"
        ):
            res = con.sql('select fastest_database_in_the_west()').fetchall()

    def test_nulls(self):
        def five_if_null(x):
            if x == None:
                return 5
            return x

        con = duckdb.connect()
        con.create_function('null_test', five_if_null, [BIGINT], BIGINT, null_handling="SPECIAL")
        res = con.sql('select null_test(NULL)').fetchall()
        assert res == [(5,)]

    @pytest.mark.parametrize(
        'pair',
        [
            (TINYINT, -129),
            (TINYINT, 128),
            (SMALLINT, -32769),
            (SMALLINT, 32768),
            (INTEGER, -2147483649),
            (INTEGER, 2147483648),
            (BIGINT, -9223372036854775815),
            (BIGINT, 9223372036854775808),
            (UTINYINT, -1),
            (UTINYINT, 256),
            (USMALLINT, -1),
            (USMALLINT, 65536),
            (UINTEGER, -1),
            (UINTEGER, 4294967296),
            (UBIGINT, -1),
            (UBIGINT, 18446744073709551616),
            (HUGEINT, -170141183460469231731687303715884105729),
            (HUGEINT, 170141183460469231731687303715884105728),
            (UHUGEINT, -1),
            (UHUGEINT, 340282366920938463463374607431768211456),
        ],
    )
    def test_return_overflow(self, pair):
        duckdb_type, overflowing_value = pair

        def return_overflow():
            return overflowing_value

        con = duckdb.connect()
        con.create_function('return_overflow', return_overflow, None, duckdb_type)
        with pytest.raises(duckdb.InvalidInputException):
            rel = con.sql('select return_overflow()')
            res = rel.fetchall()
            print(duckdb_type)
            print(res)

    def test_structs(self):
        def add_extra_column(original):
            original['a'] = 200
            original['bb'] = 0
            return original

        con = duckdb.connect()
        range_table = con.table_function('range', [5000])
        con.create_function(
            "append_field",
            add_extra_column,
            [duckdb.struct_type({'a': BIGINT, 'b': BIGINT})],
            duckdb.struct_type({'a': BIGINT, 'b': BIGINT, 'c': BIGINT}),
        )

        res = con.sql(
            """
            select append_field({'a': i::BIGINT, 'b': 3::BIGINT}) from range_table tbl(i)
        """
        )
        # added extra column to the struct
        assert len(res.fetchone()[0].keys()) == 3
        # FIXME: this is needed, otherwise the old transaction is still active when we try to start a new transaction inside of 'create_function', which means the call would fail
        res.fetchall()

        def swap_keys(dict):
            result = {}
            reversed_keys = list(dict.keys())
            reversed_keys.reverse()
            for item in reversed_keys:
                result[item] = dict[item]
            return result

        con.create_function(
            'swap_keys',
            swap_keys,
            [con.struct_type({'a': BIGINT, 'b': VARCHAR})],
            con.struct_type({'a': VARCHAR, 'b': BIGINT}),
        )
        res = con.sql(
            """
        select swap_keys({'a': 42, 'b': 'answer_to_life'})
        """
        ).fetchall()
        assert res == [({'a': 'answer_to_life', 'b': 42},)]



from .column import Column, _get_expr
from typing import Any, Callable, overload, Union

from duckdb import CaseExpression, ConstantExpression, ColumnExpression, FunctionExpression, Expression
from ._typing import ColumnOrName
from ..exception import ContributionsAcceptedError


def _invoke_function_over_columns(name: str, *cols: "ColumnOrName") -> Column:
    """
    Invokes n-ary JVM function identified by name
    and wraps the result with :class:`~pyspark.sql.Column`.
    """
    cols = [_to_column(expr) for expr in cols]
    return _invoke_function(name, *cols)


def col(column: str):
    return Column(ColumnExpression(column))


def upper(col: "ColumnOrName") -> Column:
    """
    Converts a string expression to upper case.

    .. versionadded:: 1.5.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to work on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        upper case values.

    Examples
    --------
    >>> df = spark.createDataFrame(["Spark", "PySpark", "Pandas API"], "STRING")
    >>> df.select(upper("value")).show()
    +------------+
    |upper(value)|
    +------------+
    |       SPARK|
    |     PYSPARK|
    |  PANDAS API|
    +------------+
    """
    return _invoke_function_over_columns("upper", col)


def when(condition: "Column", value: Any) -> Column:
    if not isinstance(condition, Column):
        raise TypeError("condition should be a Column")
    v = _get_expr(value)
    expr = CaseExpression(condition.expr, v)
    return Column(expr)


def _inner_expr_or_val(val):
    return val.expr if isinstance(val, Column) else val


def struct(*cols: Column) -> Column:
    return Column(FunctionExpression('struct_pack', *[_inner_expr_or_val(x) for x in cols]))


def lit(col: Any) -> Column:
    return col if isinstance(col, Column) else Column(ConstantExpression(col))


def _invoke_function(function: str, *arguments):
    return Column(FunctionExpression(function, *arguments))


def _to_column(col: ColumnOrName) -> Expression:
    return col.expr if isinstance(col, Column) else ColumnExpression(col)


def regexp_replace(str: "ColumnOrName", pattern: str, replacement: str) -> Column:
    r"""Replace all substrings of the specified string value that match regexp with rep.

    .. versionadded:: 1.5.0

    Examples
    --------
    >>> df = spark.createDataFrame([('100-200',)], ['str'])
    >>> df.select(regexp_replace('str', r'(\d+)', '--').alias('d')).collect()
    [Row(d='-----')]
    """
    return _invoke_function(
        "regexp_replace", _to_column(str), ConstantExpression(pattern), ConstantExpression(replacement), ConstantExpression('g')
    )


def array_contains(col: "ColumnOrName", value: Any) -> Column:
    """
    Collection function: returns null if the array is null, true if the array contains the
    given value, and false otherwise.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        name of column containing array
    value :
        value or column to check for in array

    Returns
    -------
    :class:`~pyspark.sql.Column`
        a column of Boolean type.

    Examples
    --------
    >>> df = spark.createDataFrame([(["a", "b", "c"],), ([],)], ['data'])
    >>> df.select(array_contains(df.data, "a")).collect()
    [Row(array_contains(data, a)=True), Row(array_contains(data, a)=False)]
    >>> df.select(array_contains(df.data, lit("a"))).collect()
    [Row(array_contains(data, a)=True), Row(array_contains(data, a)=False)]
    """
    value = _get_expr(value)
    return _invoke_function("array_contains", _to_column(col), value)


def avg(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the average of the values in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(avg(col("id"))).show()
    +-------+
    |avg(id)|
    +-------+
    |    4.5|
    +-------+
    """
    return _invoke_function_over_columns("avg", col)


def sum(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the sum of all values in the expression.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(sum(df["id"])).show()
    +-------+
    |sum(id)|
    +-------+
    |     45|
    +-------+
    """
    return _invoke_function_over_columns("sum", col)


def max(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the maximum value of the expression in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(max(col("id"))).show()
    +-------+
    |max(id)|
    +-------+
    |      9|
    +-------+
    """
    return _invoke_function_over_columns("max", col)


def mean(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the average of the values in a group.
    An alias of :func:`avg`.

    .. versionadded:: 1.4.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        the column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(mean(df.id)).show()
    +-------+
    |avg(id)|
    +-------+
    |    4.5|
    +-------+
    """
    return _invoke_function_over_columns("mean", col)


def min(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the minimum value of the expression in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    >>> df = spark.range(10)
    >>> df.select(min(df.id)).show()
    +-------+
    |min(id)|
    +-------+
    |      0|
    +-------+
    """
    return _invoke_function_over_columns("min", col)


def count(col: "ColumnOrName") -> Column:
    """
    Aggregate function: returns the number of items in a group.

    .. versionadded:: 1.3.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        target column to compute on.

    Returns
    -------
    :class:`~pyspark.sql.Column`
        column for computed results.

    Examples
    --------
    Count by all columns (start), and by a column that does not count ``None``.

    >>> df = spark.createDataFrame([(None,), ("a",), ("b",), ("c",)], schema=["alphabets"])
    >>> df.select(count(expr("*")), count(df.alphabets)).show()
    +--------+----------------+
    |count(1)|count(alphabets)|
    +--------+----------------+
    |       4|               3|
    +--------+----------------+
    """
    return _invoke_function_over_columns("count", col)


@overload
def transform(col: "ColumnOrName", f: Callable[[Column], Column]) -> Column:
    ...


@overload
def transform(col: "ColumnOrName", f: Callable[[Column, Column], Column]) -> Column:
    ...


def transform(
    col: "ColumnOrName",
    f: Union[Callable[[Column], Column], Callable[[Column, Column], Column]],
) -> Column:
    """
    Returns an array of elements after applying a transformation to each element in the input array.

    .. versionadded:: 3.1.0

    .. versionchanged:: 3.4.0
        Supports Spark Connect.

    Parameters
    ----------
    col : :class:`~pyspark.sql.Column` or str
        name of column or expression
    f : function
        a function that is applied to each element of the input array.
        Can take one of the following forms:

        - Unary ``(x: Column) -> Column: ...``
        - Binary ``(x: Column, i: Column) -> Column...``, where the second argument is
            a 0-based index of the element.

        and can use methods of :class:`~pyspark.sql.Column`, functions defined in
        :py:mod:`pyspark.sql.functions` and Scala ``UserDefinedFunctions``.
        Python ``UserDefinedFunctions`` are not supported
        (`SPARK-27052 <https://issues.apache.org/jira/browse/SPARK-27052>`__).

    Returns
    -------
    :class:`~pyspark.sql.Column`
        a new array of transformed elements.

    Examples
    --------
    >>> df = spark.createDataFrame([(1, [1, 2, 3, 4])], ("key", "values"))
    >>> df.select(transform("values", lambda x: x * 2).alias("doubled")).show()
    +------------+
    |     doubled|
    +------------+
    |[2, 4, 6, 8]|
    +------------+

    >>> def alternate(x, i):
    ...     return when(i % 2 == 0, x).otherwise(-x)
    ...
    >>> df.select(transform("values", alternate).alias("alternated")).show()
    +--------------+
    |    alternated|
    +--------------+
    |[1, -2, 3, -4]|
    +--------------+
    """
    raise NotImplementedError



import duckdb
import os
import math
from pytest import mark, fixture, importorskip

read_csv = importorskip('pyarrow.csv').read_csv
requests = importorskip('requests')
np = importorskip('numpy')


def download_file(url, name):
    r = requests.get(url, allow_redirects=True)
    open(name, 'wb').write(r.content)


def group_by_q1(con):
    con.execute("CREATE TABLE ans AS SELECT id1, sum(v1) AS v1 FROM x GROUP BY id1")
    res = con.execute("SELECT COUNT(*), sum(v1)::varchar AS v1 FROM ans").fetchall()
    assert res == [(96, '28498857')]
    con.execute("DROP TABLE ans")


def group_by_q2(con):
    con.execute("CREATE TABLE ans AS SELECT id1, id2, sum(v1) AS v1 FROM x GROUP BY id1, id2;")
    res = con.execute("SELECT count(*), sum(v1) AS v1 FROM ans;").fetchall()
    assert res == [(9216, 28498857)]
    con.execute("DROP TABLE ans")


def group_by_q3(con):
    con.execute("CREATE TABLE ans AS SELECT id3, sum(v1) AS v1, avg(v3) AS v3 FROM x GROUP BY id3;")
    res = con.execute("SELECT COUNT(*), sum(v1) AS v1, sum(v3) AS v3 FROM ans;").fetchall()
    assert res[0][0] == 95001
    assert res[0][1] == 28498857
    assert math.floor(res[0][2]) == 4749467
    con.execute("DROP TABLE ans")


def group_by_q4(con):
    con.execute("CREATE TABLE ans AS SELECT id4, avg(v1) AS v1, avg(v2) AS v2, avg(v3) AS v3 FROM x GROUP BY id4;")
    res = con.execute("SELECT COUNT(*), sum(v1) AS v1, sum(v2) AS v2, sum(v3) AS v3 FROM ans").fetchall()
    assert res[0][0] == 96
    assert math.floor(res[0][1]) == 287
    assert math.floor(res[0][2]) == 767
    assert math.floor(res[0][3]) == 4799
    con.execute("DROP TABLE ans")


def group_by_q5(con):
    con.execute("CREATE TABLE ans AS SELECT id6, sum(v1) AS v1, sum(v2) AS v2, sum(v3) AS v3 FROM x GROUP BY id6;")
    res = con.execute("SELECT COUNT(*), sum(v1) AS v1, sum(v2) AS v2, sum(v3) AS v3 FROM ans").fetchall()
    assert res[0][0] == 95001
    assert res[0][1] == 28498857
    assert res[0][2] == 75988394
    assert math.floor(res[0][3]) == 474969574
    con.execute("DROP TABLE ans")


def group_by_q6(con):
    con.execute(
        "CREATE TABLE ans AS SELECT id4, id5, quantile_cont(v3, 0.5) AS median_v3, stddev(v3) AS sd_v3 FROM x GROUP BY id4, id5;"
    )
    res = con.execute("SELECT COUNT(*), sum(median_v3) AS median_v3, sum(sd_v3) AS sd_v3 FROM ans").fetchall()
    assert res[0][0] == 9216
    assert math.floor(res[0][1]) == 460771
    assert math.floor(res[0][2]) == 266006
    con.execute("DROP TABLE ans")


def group_by_q7(con):
    con.execute("CREATE TABLE ans AS SELECT id3, max(v1)-min(v2) AS range_v1_v2 FROM x GROUP BY id3;")
    res = con.execute("SELECT count(*), sum(range_v1_v2) AS range_v1_v2 FROM ans;").fetchall()
    assert res[0][0] == 95001
    assert res[0][1] == 379850
    con.execute("DROP TABLE ans")


def group_by_q8(con):
    con.execute(
        "CREATE TABLE ans AS SELECT id6, v3 AS largest2_v3 FROM (SELECT id6, v3, row_number() OVER (PARTITION BY id6 ORDER BY v3 DESC) AS order_v3 FROM x WHERE v3 IS NOT NULL) sub_query WHERE order_v3 <= 2"
    )
    res = con.execute("SELECT count(*), sum(largest2_v3) AS largest2_v3 FROM ans").fetchall()
    assert res[0][0] == 190002
    assert math.floor(res[0][1]) == 18700554
    con.execute("DROP TABLE ans")


def group_by_q9(con):
    con.execute("CREATE TABLE ans AS SELECT id2, id4, pow(corr(v1, v2), 2) AS r2 FROM x GROUP BY id2, id4;")
    res = con.execute("SELECT count(*), sum(r2) AS r2 FROM ans").fetchall()
    assert res[0][0] == 9216
    assert math.floor(res[0][1]) == 9
    con.execute("DROP TABLE ans")


def group_by_q10(con):
    con.execute(
        "CREATE TABLE ans AS SELECT id1, id2, id3, id4, id5, id6, sum(v3) AS v3, count(*) AS count FROM x GROUP BY id1, id2, id3, id4, id5, id6;"
    )
    res = con.execute("SELECT sum(v3) AS v3, sum(count) AS count FROM ans;").fetchall()
    assert math.floor(res[0][0]) == 474969574
    assert res[0][1] == 10000000
    con.execute("DROP TABLE ans")


def join_by_q1(con):
    con.execute("CREATE TABLE ans AS SELECT x.*, small.id4 AS small_id4, v2 FROM x JOIN small USING (id1);")
    res = con.execute("SELECT COUNT(*), SUM(v1) AS v1, SUM(v2) AS v2 FROM ans;").fetchall()
    assert res[0][0] == 8998860
    assert math.floor(res[0][1]) == 450015153
    assert math.floor(res[0][2]) == 347720187
    con.execute("DROP TABLE ans")


def join_by_q2(con):
    con.execute(
        "CREATE TABLE ans AS SELECT x.*, medium.id1 AS medium_id1, medium.id4 AS medium_id4, medium.id5 AS medium_id5, v2 FROM x JOIN medium USING (id2);"
    )
    res = con.execute("SELECT COUNT(*), SUM(v1) AS v1, SUM(v2) AS v2 FROM ans;").fetchall()
    assert res[0][0] == 8998412
    assert math.floor(res[0][1]) == 449954076
    assert math.floor(res[0][2]) == 449999844
    con.execute("DROP TABLE ans")


def join_by_q3(con):
    con.execute(
        "CREATE TABLE ans AS SELECT x.*, medium.id1 AS medium_id1, medium.id4 AS medium_id4, medium.id5 AS medium_id5, v2 FROM x LEFT JOIN medium USING (id2);"
    )
    res = con.execute("SELECT COUNT(*), SUM(v1) AS v1, SUM(v2) AS v2 FROM ans;").fetchall()
    assert res[0][0] == 10000000
    assert math.floor(res[0][1]) == 500043740
    assert math.floor(res[0][2]) == 449999844
    con.execute("DROP TABLE ans")


def join_by_q4(con):
    con.execute(
        "CREATE TABLE ans AS SELECT x.*, medium.id1 AS medium_id1, medium.id2 AS medium_id2, medium.id4 AS medium_id4, v2 FROM x JOIN medium USING (id5);"
    )
    res = con.execute("SELECT COUNT(*), SUM(v1) AS v1, SUM(v2) AS v2 FROM ans;").fetchall()
    assert res[0][0] == 8998412
    assert math.floor(res[0][1]) == 449954076
    assert math.floor(res[0][2]) == 449999844
    con.execute("DROP TABLE ans")


def join_by_q5(con):
    con.execute(
        "CREATE TABLE ans AS SELECT x.*, big.id1 AS big_id1, big.id2 AS big_id2, big.id4 AS big_id4, big.id5 AS big_id5, big.id6 AS big_id6, v2 FROM x JOIN big USING (id3);"
    )
    res = con.execute("SELECT COUNT(*), SUM(v1) AS v1, SUM(v2) AS v2 FROM ans;").fetchall()
    assert res[0][0] == 9000000
    assert math.floor(res[0][1]) == 450032091
    assert math.floor(res[0][2]) == 449860428
    con.execute("DROP TABLE ans")


class TestH2OAIArrow(object):
    @mark.parametrize(
        'function',
        [
            group_by_q1,
            group_by_q2,
            group_by_q3,
            group_by_q4,
            group_by_q5,
            group_by_q6,
            group_by_q7,
            group_by_q8,
            group_by_q9,
            group_by_q10,
        ],
    )
    @mark.parametrize('threads', [1, 4])
    def test_group_by(self, threads, function):
        con = duckdb.connect()
        download_file(
            'https://github.com/duckdb/duckdb-data/releases/download/v1.0/G1_1e7_1e2_5_0.csv.gz',
            'G1_1e7_1e2_5_0.csv.gz',
        )
        arrow_table = read_csv('G1_1e7_1e2_5_0.csv.gz')
        con.register("x", arrow_table)
        os.remove('G1_1e7_1e2_5_0.csv.gz')

        con.execute(f"PRAGMA threads={threads}")

        function(con)

    @mark.parametrize('threads', [1, 4])
    @mark.parametrize(
        'function',
        [
            join_by_q1,
            join_by_q2,
            join_by_q3,
            join_by_q4,
            join_by_q5,
        ],
    )
    @mark.usefixtures('large_data')
    def test_join(self, threads, function, large_data):
        large_data.execute(f"PRAGMA threads={threads}")

        function(large_data)


@fixture(scope="module")
def large_data():
    download_file(
        'https://github.com/duckdb/duckdb-data/releases/download/v1.0/J1_1e7_NA_0_0.csv.gz', 'J1_1e7_NA_0_0.csv.gz'
    )
    download_file(
        'https://github.com/duckdb/duckdb-data/releases/download/v1.0/J1_1e7_1e1_0_0.csv.gz', 'J1_1e7_1e1_0_0.csv.gz'
    )
    download_file(
        'https://github.com/duckdb/duckdb-data/releases/download/v1.0/J1_1e7_1e4_0_0.csv.gz', 'J1_1e7_1e4_0_0.csv.gz'
    )
    download_file(
        'https://github.com/duckdb/duckdb-data/releases/download/v1.0/J1_1e7_1e7_0_0.csv.gz', 'J1_1e7_1e7_0_0.csv.gz'
    )

    con = duckdb.connect()
    arrow_table = read_csv('J1_1e7_NA_0_0.csv.gz')
    con.register("x", arrow_table)

    arrow_table = read_csv('J1_1e7_1e1_0_0.csv.gz')
    con.register("small", arrow_table)

    arrow_table = read_csv('J1_1e7_1e4_0_0.csv.gz')
    con.register("medium", arrow_table)

    arrow_table = read_csv('J1_1e7_1e7_0_0.csv.gz')
    con.register("big", arrow_table)

    yield con

    os.remove('J1_1e7_NA_0_0.csv.gz')
    os.remove('J1_1e7_1e1_0_0.csv.gz')
    os.remove('J1_1e7_1e4_0_0.csv.gz')
    os.remove('J1_1e7_1e7_0_0.csv.gz')



import numpy as np
import datetime
import duckdb
import pytest


class TestPandasObjectInteger(object):
    # Signed Masked Integer types
    def test_object_integer(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        df_in = pd.DataFrame(
            {
                'int8': pd.Series([None, 1, -1], dtype="Int8"),
                'int16': pd.Series([None, 1, -1], dtype="Int16"),
                'int32': pd.Series([None, 1, -1], dtype="Int32"),
                'int64': pd.Series([None, 1, -1], dtype="Int64"),
            }
        )
        # These are float64 because pandas would force these to be float64 even if we set them to int8, int16, int32, int64 respectively
        df_expected_res = pd.DataFrame(
            {
                'int8': np.ma.masked_array([0, 1, -1], mask=[True, False, False], dtype='float64'),
                'int16': np.ma.masked_array([0, 1, -1], mask=[True, False, False], dtype='float64'),
                'int32': np.ma.masked_array([0, 1, -1], mask=[True, False, False], dtype='float64'),
                'int64': np.ma.masked_array([0, 1, -1], mask=[True, False, False], dtype='float64'),
            }
        )
        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        pd.testing.assert_frame_equal(df_expected_res, df_out)

    # Unsigned Masked Integer types
    def test_object_uinteger(self, duckdb_cursor):
        pd = pytest.importorskip("pandas")
        df_in = pd.DataFrame(
            {
                'uint8': pd.Series([None, 1, 255], dtype="UInt8"),
                'uint16': pd.Series([None, 1, 65535], dtype="UInt16"),
                'uint32': pd.Series([None, 1, 4294967295], dtype="UInt32"),
                'uint64': pd.Series([None, 1, 18446744073709551615], dtype="UInt64"),
            }
        )
        # These are float64 because pandas would force these to be float64 even if we set them to uint8, uint16, uint32, uint64 respectively
        df_expected_res = pd.DataFrame(
            {
                'uint8': np.ma.masked_array([0, 1, 255], mask=[True, False, False], dtype='float64'),
                'uint16': np.ma.masked_array([0, 1, 65535], mask=[True, False, False], dtype='float64'),
                'uint32': np.ma.masked_array([0, 1, 4294967295], mask=[True, False, False], dtype='float64'),
                'uint64': np.ma.masked_array([0, 1, 18446744073709551615], mask=[True, False, False], dtype='float64'),
            }
        )
        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        pd.testing.assert_frame_equal(df_expected_res, df_out)

    # Unsigned Masked float/double types
    def test_object_float(self, duckdb_cursor):
        # Require pandas 1.2.0 >= for this, because Float32|Float64 was not added before this version
        pd = pytest.importorskip("pandas", '1.2.0')
        df_in = pd.DataFrame(
            {
                'float32': pd.Series([None, 1, 4294967295], dtype="Float32"),
                'float64': pd.Series([None, 1, 18446744073709551615], dtype="Float64"),
            }
        )
        df_expected_res = pd.DataFrame(
            {
                'float32': np.ma.masked_array([0, 1, 4294967295], mask=[True, False, False], dtype='float32'),
                'float64': np.ma.masked_array([0, 1, 18446744073709551615], mask=[True, False, False], dtype='float64'),
            }
        )
        df_out = duckdb.query_df(df_in, "data", "SELECT * FROM data").df()
        pd.testing.assert_frame_equal(df_expected_res, df_out)



import pytest

_ = pytest.importorskip("duckdb.experimental.spark")

from duckdb.experimental.spark.sql.types import (
    LongType,
    StructType,
    BooleanType,
    StructField,
    StringType,
    IntegerType,
    LongType,
    Row,
    ArrayType,
    MapType,
)
from duckdb.experimental.spark.sql.functions import col, struct, when, lit
import duckdb
import re


class TestWithColumnRenamed(object):
    def test_with_column_renamed(self, spark):
        dataDF = [
            (('James', '', 'Smith'), '1991-04-01', 'M', 3000),
            (('Michael', 'Rose', ''), '2000-05-19', 'M', 4000),
            (('Robert', '', 'Williams'), '1978-09-05', 'M', 4000),
            (('Maria', 'Anne', 'Jones'), '1967-12-01', 'F', 4000),
            (('Jen', 'Mary', 'Brown'), '1980-02-17', 'F', -1),
        ]
        from duckdb.experimental.spark.sql.types import StructType, StructField, StringType, IntegerType

        schema = StructType(
            [
                StructField(
                    'name',
                    StructType(
                        [
                            StructField('firstname', StringType(), True),
                            StructField('middlename', StringType(), True),
                            StructField('lastname', StringType(), True),
                        ]
                    ),
                ),
                StructField('dob', StringType(), True),
                StructField('gender', StringType(), True),
                StructField('salary', IntegerType(), True),
            ]
        )

        df = spark.createDataFrame(data=dataDF, schema=schema)

        df2 = df.withColumnRenamed("dob", "DateOfBirth").withColumnRenamed("salary", "salary_amount")
        assert 'dob' not in df2
        assert 'salary' not in df2
        assert 'DateOfBirth' in df2
        assert 'salary_amount' in df2

        schema2 = StructType(
            [
                StructField(
                    'full name',
                    StructType(
                        [
                            StructField('fname', StringType(), True),
                            StructField('mname', StringType(), True),
                            StructField('lname', StringType(), True),
                        ]
                    ),
                ),
            ]
        )

        df2 = df.withColumnRenamed("name", "full name")
        assert 'name' not in df2
        assert 'full name' in df2
        assert 'firstname' in df2.schema['full name'].dataType

        df2 = df.select(
            col("name").alias("full name"),
            col("dob"),
            col("gender"),
            col("salary"),
        )
        assert 'name' not in df2
        assert 'full name' in df2
        assert 'firstname' in df2.schema['full name'].dataType

        df2 = df.select(
            col("name.firstname").alias("fname"),
            col("name.middlename").alias("mname"),
            col("name.lastname").alias("lname"),
            col("dob"),
            col("gender"),
            col("salary"),
        )
        assert 'firstname' not in df2
        assert 'fname' in df2



import duckdb
import pytest
import datetime
import pytz

pa = pytest.importorskip('pyarrow')


def generate_table(current_time, precision, timezone):
    timestamp_type = pa.timestamp(precision, tz=timezone)
    schema = pa.schema([("data", timestamp_type)])
    inputs = [pa.array([current_time], type=timestamp_type)]
    return pa.Table.from_arrays(inputs, schema=schema)


timezones = ['UTC', 'BET', 'CET', 'Asia/Kathmandu']


class TestArrowTimestampsTimezone(object):
    def test_timestamp_timezone(self, duckdb_cursor):
        precisions = ['us', 's', 'ns', 'ms']
        current_time = datetime.datetime(2017, 11, 28, 23, 55, 59, tzinfo=pytz.UTC)
        con = duckdb.connect()
        con.execute("SET TimeZone = 'UTC'")
        for precision in precisions:
            arrow_table = generate_table(current_time, precision, 'UTC')
            res_utc = con.from_arrow(arrow_table).execute().fetchall()
            assert res_utc[0][0] == current_time

    def test_timestamp_timezone_overflow(self, duckdb_cursor):
        precisions = ['s', 'ms']
        current_time = 9223372036854775807
        for precision in precisions:
            with pytest.raises(duckdb.ConversionException, match='Could not convert'):
                arrow_table = generate_table(current_time, precision, 'UTC')
                res_utc = duckdb.from_arrow(arrow_table).execute().fetchall()

    def test_timestamp_tz_to_arrow(self, duckdb_cursor):
        precisions = ['us', 's', 'ns', 'ms']
        current_time = datetime.datetime(2017, 11, 28, 23, 55, 59)
        con = duckdb.connect()
        for precision in precisions:
            for timezone in timezones:
                con.execute("SET TimeZone = '" + timezone + "'")
                arrow_table = generate_table(current_time, precision, timezone)
                res = con.from_arrow(arrow_table).arrow()
                assert res[0].type == pa.timestamp('us', tz=timezone)
                assert res == generate_table(current_time, 'us', timezone)

    def test_timestamp_tz_with_null(self, duckdb_cursor):
        con = duckdb.connect()
        con.execute("create table t (i timestamptz)")
        con.execute("insert into t values (NULL),('2021-11-15 02:30:00'::timestamptz)")
        rel = con.table('t')
        arrow_tbl = rel.arrow()
        con.register('t2', arrow_tbl)

        assert con.execute("select * from t").fetchall() == con.execute("select * from t2").fetchall()

    def test_timestamp_stream(self, duckdb_cursor):
        con = duckdb.connect()
        con.execute("create table t (i timestamptz)")
        con.execute("insert into t values (NULL),('2021-11-15 02:30:00'::timestamptz)")
        rel = con.table('t')
        arrow_tbl = rel.record_batch().read_all()
        con.register('t2', arrow_tbl)

        assert con.execute("select * from t").fetchall() == con.execute("select * from t2").fetchall()



from .. import DuckDBPyConnection

from typing import Dict

SPECIAL: FunctionNullHandling
DEFAULT: FunctionNullHandling

NATIVE: PythonUDFType
ARROW: PythonUDFType

class FunctionNullHandling:
    DEFAULT: FunctionNullHandling
    SPECIAL: FunctionNullHandling
    def __int__(self) -> int: ...
    def __index__(self) -> int: ...
    @property
    def __members__(self) -> Dict[str, FunctionNullHandling]: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...

class PythonUDFType:
    NATIVE: PythonUDFType
    ARROW: PythonUDFType
    def __int__(self) -> int: ...
    def __index__(self) -> int: ...
    @property
    def __members__(self) -> Dict[str, PythonUDFType]: ...
    @property
    def name(self) -> str: ...
    @property
    def value(self) -> int: ...
